###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" Collection of user jobs for testing purposes
"""
import os
import time
import errno

from DIRAC import rootPath
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.tests.Utilities.testJobDefinitions import baseToAllJobs, endOfAllJobs, find_all
from DIRAC.Core.Utilities.Proxy import executeWithUserProxy
from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob
from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb

# parameters

jobClass = LHCbJob
diracClass = DiracLHCb

try:
    tier1s = DMSHelpers().getTiers(tier=(0, 1))
except AttributeError:
    tier1s = [
        "LCG.CERN.cern",
        "LCG.CNAF.it",
        "LCG.GRIDKA.de",
        "LCG.IN2P3.fr",
        "LCG.NIKHEF.nl",
        "LCG.PIC.es",
        "LCG.RAL.uk",
        "LCG.NCBJ.pl",
        "LCG.SARA.nl",
    ]

# List of jobs
wdir = os.getcwd()


@executeWithUserProxy
def helloWorldTestT2s():
    job = baseToAllJobs("helloWorldTestT2s", jobClass)
    job.setInputSandbox([find_all("exe-script.py", rootPath, ".")[0]])
    job.setExecutable("exe-script.py", "", "helloWorld.log")
    job.setBannedSites(tier1s)
    return endOfAllJobs(job)


@executeWithUserProxy
def helloWorldTestCERN():
    job = baseToAllJobs("helloWorld-test-CERN", jobClass)
    job.setInputSandbox([find_all("exe-script.py", rootPath, ".")[0]])
    job.setExecutable("exe-script.py", "", "helloWorld.log")
    job.setDestination("LCG.CERN.cern")
    return endOfAllJobs(job)


@executeWithUserProxy
def helloWorldTestIN2P3():
    job = baseToAllJobs("helloWorld-test-IN2P3", jobClass)
    job.setInputSandbox([find_all("exe-script.py", rootPath, ".")[0]])
    job.setExecutable("exe-script.py", "", "helloWorld.log")
    job.setDestination("LCG.IN2P3.fr")
    return endOfAllJobs(job)


@executeWithUserProxy
def helloWorldTestGRIDKA():
    job = baseToAllJobs("helloWorld-test-GRIDKA", jobClass)
    job.setInputSandbox([find_all("exe-script.py", rootPath, ".")[0]])
    job.setExecutable("exe-script.py", "", "helloWorld.log")
    job.setDestination("LCG.GRIDKA.de")
    return endOfAllJobs(job)


@executeWithUserProxy
def helloWorldTestARC():
    job = baseToAllJobs("helloWorld-test-ARC", jobClass)
    job.setInputSandbox([find_all("exe-script.py", rootPath, ".")[0]])
    job.setExecutable("exe-script.py", "", "helloWorld.log")
    job.setDestination(["LCG.RAL.uk"])
    return endOfAllJobs(job)


@executeWithUserProxy
def helloWorldTestSSHCondor():
    job = baseToAllJobs("helloWorld-test-SSHCondor", jobClass)
    job.setInputSandbox([find_all("exe-script.py", rootPath, ".")[0]])
    job.setExecutable("exe-script.py", "", "helloWorld.log")
    job.setDestination(["DIRAC.Sibir.ru"])
    return endOfAllJobs(job)


@executeWithUserProxy
def helloWorldTestARM():
    job = baseToAllJobs("helloWorld-test-ARM", jobClass)
    job.setInputSandbox([find_all("exe-script.py", rootPath, ".")[0]])
    job.setExecutable("exe-script.py", "", "helloWorld.log")
    job.setDestination(["DIRAC.ARM.ch"])
    return endOfAllJobs(job)


@executeWithUserProxy
def jobWithOutput():
    timenow = time.strftime("%s")
    with open(os.path.join(wdir, timenow + "testFileUpload.txt"), "w") as f:
        f.write(timenow)
    inp1 = [find_all(timenow + "testFileUpload.txt", ".")[0]]
    inp2 = [find_all("exe-script.py", rootPath, ".")[0]]
    job = baseToAllJobs("jobWithOutput", jobClass)
    job.setInputSandbox(inp1 + inp2)
    job.setExecutable("exe-script.py", "", "helloWorld.log")
    job.setOutputData([timenow + "testFileUpload.txt"])
    res = endOfAllJobs(job)
    try:
        os.remove(os.path.join(wdir, timenow + "testFileUpload.txt"))
    except OSError as e:
        return e.errno == errno.ENOENT
    return res


@executeWithUserProxy
def jobWithOutputAndPrepend():
    timenow = time.strftime("%s")
    with open(os.path.join(wdir, timenow + "testFileUploadNewPath.txt"), "w") as f:
        f.write(timenow)
    job = baseToAllJobs("jobWithOutputAndPrepend", jobClass)
    inp1 = [find_all(timenow + "testFileUploadNewPath.txt", ".")[0]]
    inp2 = [find_all("exe-script.py", rootPath, ".")[0]]
    job.setInputSandbox(inp1 + inp2)
    job.setExecutable("exe-script.py", "", "helloWorld.log")
    job.setOutputData([timenow + "testFileUploadNewPath.txt"], filePrepend="testFilePrepend")
    res = endOfAllJobs(job)
    try:
        os.remove(os.path.join(wdir, timenow + "testFileUploadNewPath.txt"))
    except OSError as e:
        return e.errno == errno.ENOENT
    return res


@executeWithUserProxy
def jobWithOutputAndPrependWithUnderscore():
    timenow = time.strftime("%s")
    with open(os.path.join(wdir, timenow + "testFileUpload_NewPath.txt"), "w") as f:
        f.write(timenow)
    job = baseToAllJobs("jobWithOutputAndPrependWithUnderscore", jobClass)
    inp1 = [find_all(timenow + "testFileUpload_NewPath.txt", ".")[0]]
    inp2 = [find_all("exe-script.py", rootPath, ".")[0]]
    job.setInputSandbox(inp1 + inp2)
    job.setExecutable("exe-script.py", "", "helloWorld.log")
    res = job.setOutputData([timenow + "testFileUpload_NewPath.txt"], filePrepend="testFilePrepend")
    if not res["OK"]:
        return 0
    res = endOfAllJobs(job)
    try:
        os.remove(os.path.join(wdir, timenow + "testFileUpload_NewPath.txt"))
    except OSError as e:
        return e.errno == errno.ENOENT
    return res


@executeWithUserProxy
def jobWithOutputAndReplication():
    timenow = time.strftime("%s")
    with open(os.path.join(wdir, timenow + "testFileReplication.txt"), "w") as f:
        f.write(timenow)
    job = baseToAllJobs("jobWithOutputAndReplication", jobClass)
    inp1 = [find_all(timenow + "testFileReplication.txt", ".")[0]]
    inp2 = [find_all("exe-script.py", rootPath, ".")[0]]
    job.setInputSandbox(inp1 + inp2)
    job.setExecutable("exe-script.py", "", "helloWorld.log")
    job.setOutputData([timenow + "testFileReplication.txt"], replicate="True")
    res = endOfAllJobs(job)
    try:
        os.remove(os.path.join(wdir, timenow + "testFileReplication.txt"))
    except OSError as e:
        return e.errno == errno.ENOENT
    return res


@executeWithUserProxy
def jobWith2OutputsToBannedSE():
    timenow = time.strftime("%s")
    with open(os.path.join(wdir, timenow + "testFileUploadBanned-1.txt"), "w") as f:
        f.write(timenow)
    with open(os.path.join(wdir, timenow + "testFileUploadBanned-2.txt"), "w") as f:
        f.write(timenow)
    job = baseToAllJobs("jobWith2OutputsToBannedSE", jobClass)
    inp1 = [find_all(timenow + "testFileUploadBanned-1.txt", ".")[0]]
    inp2 = [find_all(timenow + "testFileUploadBanned-2.txt", ".")[0]]
    inp3 = [find_all("exe-script.py", rootPath, ".")[0]]
    inp4 = [find_all("partialConfig.cfg", "..", ".")[0]]
    job.setInputSandbox(inp1 + inp2 + inp3 + inp4)
    job.setExecutable("exe-script.py", "", "helloWorld.log")
    job.setConfigArgs("partialConfig.cfg")
    job.setDestination("LCG.PIC.es")
    job.setOutputData(
        [timenow + "testFileUploadBanned-1.txt", timenow + "testFileUploadBanned-2.txt"], OutputSE=["PIC-USER"]
    )
    res = endOfAllJobs(job)
    try:
        os.remove(os.path.join(wdir, timenow + "testFileUploadBanned-1.txt"))
    except OSError as e:
        return e.errno == errno.ENOENT
    try:
        os.remove(os.path.join(wdir, timenow + "testFileUploadBanned-2.txt"))
    except OSError as e:
        return e.errno == errno.ENOENT
    return res


@executeWithUserProxy
def jobWithSingleInputData():
    job = baseToAllJobs("jobWithSingleInputData-shouldGoToCERN", jobClass)
    job.setInputSandbox([find_all("exe-script-with-input-single-location.py", "..", ".")[0]])
    job.setExecutable("exe-script-with-input-single-location.py", "", "exeWithInput.log")
    # this file should be at CERN-USER only
    job.setInputData("/lhcb/user/f/fstagni/test/testInputFileSingleLocation.txt")
    job.setInputDataPolicy("download")
    res = endOfAllJobs(job)
    return res


@executeWithUserProxy
def jobWithSingleInputDataCERN():
    job = baseToAllJobs("jobWithSingleInputDataCERN-shouldSucceed", jobClass)
    job.setInputSandbox([find_all("exe-script-with-input-single-location.py", "..", ".")[0]])
    job.setExecutable("exe-script-with-input-single-location.py", "", "exeWithInput.log")
    # this file should be at CERN-USER only
    job.setInputData("/lhcb/user/f/fstagni/test/testInputFileSingleLocation.txt")
    job.setInputDataPolicy("download")
    job.setDestination(["LCG.CERN.cern"])
    res = endOfAllJobs(job)
    return res


@executeWithUserProxy
def jobWithSingleInputDataRAL():
    job = baseToAllJobs("jobWithSingleInputDataRAL-shouldFailOptimizers", jobClass)
    job.setInputSandbox([find_all("exe-script-with-input-single-location.py", "..", ".")[0]])
    job.setExecutable("exe-script-with-input-single-location.py", "", "exeWithInput.log")
    # this file should be at CERN-USER only
    job.setInputData("/lhcb/user/f/fstagni/test/testInputFileSingleLocation.txt")
    job.setInputDataPolicy("protocol")
    job.setDestination(["LCG.RAL.uk"])
    res = endOfAllJobs(job)
    return res


@executeWithUserProxy
def jobWithSingleInputDataIN2P3():
    job = baseToAllJobs("jobWithSingleInputDataIN2P3-shouldFailOptimizers", jobClass)
    job.setInputSandbox([find_all("exe-script-with-input-single-location.py", "..", ".")[0]])
    job.setExecutable("exe-script-with-input-single-location.py", "", "exeWithInput.log")
    # this file should be at CERN-USER only
    job.setInputData("/lhcb/user/f/fstagni/test/testInputFileSingleLocation.txt")
    job.setInputDataPolicy("protocol")
    job.setDestination(["LCG.IN2P3.fr"])
    res = endOfAllJobs(job)
    return res


@executeWithUserProxy
def jobWithSingleInputDataNCBJ():
    job = baseToAllJobs("jobWithSingleInputDataNCBJ-shouldFailOptimizers", jobClass)
    job.setInputSandbox([find_all("exe-script-with-input-single-location.py", "..", ".")[0]])
    job.setExecutable("exe-script-with-input-single-location.py", "", "exeWithInput.log")
    # this file should be at CERN-USER only
    job.setInputData("/lhcb/user/f/fstagni/test/testInputFileSingleLocation.txt")
    job.setInputDataPolicy("protocol")
    job.setDestination(["LCG.NCBJ.pl"])
    res = endOfAllJobs(job)
    return res


@executeWithUserProxy
def jobWithSingleInputDataSARA():
    job = baseToAllJobs("jobWithSingleInputDataSARA-shouldFailOptimizers", jobClass)
    job.setInputSandbox([find_all("exe-script-with-input-single-location.py", "..", ".")[0]])
    job.setExecutable("exe-script-with-input-single-location.py", "", "exeWithInput.log")
    # this file should be at CERN-USER only
    job.setInputData("/lhcb/user/f/fstagni/test/testInputFileSingleLocation.txt")
    job.setInputDataPolicy("protocol")
    job.setDestination(["LCG.SARA.nl"])
    res = endOfAllJobs(job)
    return res


@executeWithUserProxy
def jobWithSingleInputDataPIC():
    job = baseToAllJobs("jobWithSingleInputDataPIC-shouldFailOptimizers", jobClass)
    job.setInputSandbox([find_all("exe-script-with-input-single-location.py", "..", ".")[0]])
    job.setExecutable("exe-script-with-input-single-location.py", "", "exeWithInput.log")
    # this file should be at CERN-USER only
    job.setInputData("/lhcb/user/f/fstagni/test/testInputFileSingleLocation.txt")
    job.setInputDataPolicy("protocol")
    job.setDestination(["LCG.PIC.es"])
    res = endOfAllJobs(job)
    return res


@executeWithUserProxy
def jobWithSingleInputDataSpreaded():
    job = baseToAllJobs("jobWithSingleInputDataSpreaded", jobClass)
    job.setInputSandbox([find_all("exe-script-with-input.py", "..", ".")[0]])
    job.setExecutable("exe-script-with-input.py", "", "exeWithInput.log")
    # this file should be at CERN-USER and IN2P3-USER
    job.setInputData("/lhcb/user/f/fstagni/test/testInputFile.txt")
    job.setInputDataPolicy("download")
    res = endOfAllJobs(job)
    return res


@executeWithUserProxy
def jobWithInputDataAndAncestor():
    job = baseToAllJobs("jobWithInputDataAndAncestor", jobClass)
    job.setInputSandbox([find_all("exe-script-with-input-and-ancestor.py", "..", ".")[0]])
    job.setExecutable("exe-script-with-input-and-ancestor.py", "", "exeWithInput.log")
    # WARNING: Collision10!!
    job.setInputData("/lhcb/data/2010/SDST/00008375/0005/00008375_00053941_1.sdst")  # this file should be at SARA-RDST
    # the ancestor should be /lhcb/data/2010/RAW/FULL/LHCb/COLLISION10/81616/081616_0000000213.raw (CERN and SARA)
    job.setAncestorDepth(1)  # pylint: disable=no-member
    job.setInputDataPolicy("download")
    res = endOfAllJobs(job)
    return res


@executeWithUserProxy
def gaussMPJob():
    job = baseToAllJobs("GaussMP_v54r3", jobClass)

    job.setInputSandbox([find_all("prodConf_Gauss_MP_test.py", "..", ".")[0]])
    job.setOutputSandbox("Gauss_MP_test.sim")

    options = "$APPCONFIGOPTS/Gauss/Beam7000GeV-mu100-nu7.6-HorExtAngle.py;"
    options += "$APPCONFIGOPTS/Gauss/EnableSpillover-25ns.py;"
    options += "$DECFILESROOT/options/12143001.py;"
    options += "$LBPYTHIA8ROOT/options/Pythia8.py;"
    options += "$APPCONFIGOPTS/Gauss/Gauss-Upgrade-Baseline-20150522.py;"
    options += "$APPCONFIGOPTS/Gauss/G4PL_FTFP_BERT_EmOpt2.py;"
    options += "$APPCONFIGOPTS/Gauss/GaussMPpatch20200701.py;"
    options += "$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py"

    job.setApplication(
        "Gauss",
        "v54r3",
        options,  # pylint: disable=no-member
        extraPackages="AppConfig.v3r400;Gen/DecFiles.v30r42;ProdConf.v3r0",
        systemConfig="x86_64-centos7-gcc9-opt",
        eventTimeout=7200,
    )
    job.setDIRACPlatform()  # pylint: disable=no-member
    job.setCPUTime(172800)
    job.setNumberOfProcessors(4)

    return endOfAllJobs(job)



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set the status (default Unused) of a list of LFNs or files in status
<Status> of Transformation <TransID>"""
import DIRAC

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()
    newStatus = "Unused"
    statusList = (
        "Unused",
        "Assigned",
        "Done",
        "Problematic",
        "MissingLFC",
        "MissingInFC",
        "MaxReset",
        "Processed",
        "NotProcessed",
        "Removed",
        "ProbInFC",
    )
    Script.registerSwitch("", "Status=", f"Select files with a given status from {str(statusList)}")
    Script.registerSwitch("", "NewStatus=", f"New status to be set (default: {newStatus})")
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] <TransID> <Status>",
            ]
        )
    )
    Script.parseCommandLine()

    from DIRAC.Core.Utilities.List import breakListIntoChunks
    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
    from DIRAC.TransformationSystem.Utilities.ScriptUtilities import getTransformations

    switches = Script.getUnprocessedSwitches()
    status = None
    for opt, val in switches:
        if opt == "Status":
            val = set(val.split(","))
            if val & set(statusList) != val:
                print(f"Unknown status {','.join(val)}... Select in {str(statusList)}")
                Script.showHelp(exitCode=1)
            status = list(val)
        elif opt == "NewStatus":
            if val not in statusList:
                print(f"Unknown status {val}... Select in {str(statusList)}")
                Script.showHelp()
                DIRAC.exit(1)
            newStatus = val

    args = Script.getPositionalArgs()
    idList = getTransformations([args[0]])
    if not idList:
        DIRAC.exit(1)

    if len(args) == 2:
        status = args[1].split(",")
    elif not status:
        status = ["Unknown"]
    lfnsExplicit = dmScript.getOption("LFNs")

    transClient = TransformationClient()

    for transID in idList:
        lfns = lfnsExplicit
        if not lfns:
            res = transClient.getTransformation(transID)
            if not res["OK"]:
                print(f"Failed to get transformation information: {res['Message']}")
                DIRAC.exit(2)

            selectDict = {"TransformationID": res["Value"]["TransformationID"], "Status": status}
            res = transClient.getTransformationFiles(condDict=selectDict)
            if not res["OK"]:
                print(f"Failed to get files: {res['Message']}")
                DIRAC.exit(2)

            lfns = [d["LFN"] for d in res["Value"]]
            if not lfns:
                print(f"No files found in transformation {transID}, status {status}")

        if not lfns:
            print("No files to be set in transformation", transID)
        else:
            resetFiles = 0
            failed = {}
            for lfnChunk in breakListIntoChunks(lfns, 10000):
                force = "MaxReset" in status or "Processed" in status or lfnsExplicit
                res = transClient.setFileStatusForTransformation(transID, newStatus, lfnChunk, force=force)
                if res["OK"]:
                    resetFiles += len(res["Value"].get("Successful", res["Value"]))
                    for lfn, reason in res["Value"].get("Failed", {}).items():
                        if reason != "File not found in the Transformation Database":
                            failed.setdefault(reason, []).append(lfn)
                else:
                    print(
                        "Failed to set %d files to %s in transformation %s: %s"
                        % (len(lfns), newStatus, transID, res["Message"])
                    )
            print("%d files were set %s in transformation %s" % (resetFiles, newStatus, transID))
            if failed:
                for reason in failed:
                    print(f"Failed for {len(failed[reason])} files: {reason}")
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set Data Quality Flag for the given run."""
import DIRAC
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... Run Flag",
                "Arguments:",
                "  Run:      Run number",
                "  Flag:     Quality Flag",
            ]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    bk = BookkeepingClient()

    if len(args) < 2:
        result = bk.getAvailableDataQuality()
        if not result["OK"]:
            print(f"ERROR: {result['Message']}")
            DIRAC.exit(2)
        flags = result["Value"]
        print("Available Data Quality Flags")
        for flag in flags:
            print(flag)
        Script.showHelp()

    exitCode = 0
    rnb = int(args[0])
    flag = args[1]
    result = bk.setRunDataQuality(rnb, flag)

    if not result["OK"]:
        print(f"ERROR: {result['Message']}")
        exitCode = 2
    else:
        succ = result["Value"]["Successful"]
        failed = result["Value"]["Failed"]
        print("The data quality has been set for the following files:")
        for i in succ:
            print(i)

        if len(failed) != 0:
            print("The data quality has not been set for the following files:")
            for i in failed:
                print(i)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Check if the files are in the BK, the FC and the SEs they are supposed to be
in.

Uses the DM script switches, and, unless a list of LFNs is provided

1) If --Directory is used: get files in FC directories
2) If --Production or --BK options is used get files in the FC directories from the BK

If an (list of) SE or SE-group is specified option --SE) and the file is not in the FC,
check if it is in these SEs

- If the file is in the SE and in the BK, it can be registered in the FC using --FixIt

If --FixIt is set, takes actions:
  Missing files: remove from SE and FC
  No replica flag: set it (in the BK)
  Not existing in SE: remove replica or file from the catalog
  Existing in SE and BK: register it in the FC
  Bad checksum: remove replica or file from SE and catalogs if no good replica
"""

from DIRAC.Core.Base.Script import Script


def __getSEsFromOptions(dmScript):
    from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers

    seList = dmScript.getOption("SEs", [])
    sites = dmScript.getOption("Sites", [])
    if sites:
        siteSEs = []
        dmsHelper = DMSHelpers()
        for site in sites:
            siteSEs += dmsHelper.getSEsForSite(site).get("Value", [])
        if seList and siteSEs:
            seList = list(set(seList) & set(siteSEs))
        else:
            seList += siteSEs
    return seList


@Script()
def main():
    # Script initialization
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] [values]",
            ]
        )
    )
    dmScript = DMScript()
    dmScript.registerDMSwitches()
    maxFiles = 20
    Script.registerSwitch("", "FixIt", "   Take action to fix the catalogs and storage")
    Script.registerSwitch("", "Replace", "   Replace bad or missing replicas (default=False)")
    Script.registerSwitch("", "NoBK", "   Do not check with BK")
    Script.registerSwitch("", "Verbose", "   Set logging mode to INFO")
    Script.registerSwitch("", "MaxFiles=", "   Set maximum number of files to be printed (default %d)" % maxFiles)
    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import gLogger

    fixIt = False
    bkCheck = True
    replace = False
    verbose = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "FixIt":
            fixIt = True
        elif switch[0] == "NoBK":
            bkCheck = False
        elif switch[0] == "Replace":
            replace = True
        elif switch[0] == "Verbose":
            verbose = True
        elif switch[0] == "MaxFiles":
            try:
                maxFiles = int(switch[1])
            except Exception as e:
                gLogger.exception("Invalid value for MaxFiles", lException=e)
                pass

    # imports
    from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks

    if verbose:
        gLogger.setLevel("INFO")
    cc = ConsistencyChecks()
    cc.directories = dmScript.getOption("Directory", [])
    cc.lfns = dmScript.getOption("LFNs", []) + [lfn for arg in Script.getPositionalArgs() for lfn in arg.split(",")]
    bkQuery = dmScript.getBKQuery(visible="All")
    if bkQuery:
        bkQuery.setOption("ReplicaFlag", "All")
        cc.bkQuery = bkQuery

    cc.seList = __getSEsFromOptions(dmScript)
    from LHCbDIRAC.DataManagementSystem.Client.CheckExecutors import doCheckFC2SE

    try:
        doCheckFC2SE(cc, bkCheck=bkCheck, fixIt=fixIt, replace=replace, maxFiles=maxFiles)
    except RuntimeError as e:
        gLogger.fatal(str(e))
    except Exception as e:
        gLogger.exception("Exception", lException=e)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Remove replicas of a (list of) LFNs at a list of sites.

It is possible to request a minimum of remaining replicas
"""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()
    dmScript.registerSiteSwitches()

    Script.registerSwitch("n", "NoFC", " use this option to force the removal from storage of replicas not in FC")
    Script.registerSwitch(
        "", "ReduceReplicas=", "  specify the number of replicas you want to keep (default SE: Tier1-USER)"
    )
    Script.registerSwitch("", "Force", " use this option for force the removal of replicas even if last one")
    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... [LFN[,LFN2[,LFN3...]]] SE[,SE2...]",
                "Arguments:",
                "  LFN:      Logical File Name or file containing LFNs",
                "  SE:       Valid DIRAC SE",
            ]
        )
    )
    Script.parseCommandLine()

    from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeRemoveReplicas
    from DIRAC import exit

    exit(executeRemoveReplicas(dmScript))


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Interacts with pool xml catalog."""
from DIRAC.Resources.Catalog.PoolXMLCatalog import PoolXMLCatalog
from DIRAC.Resources.Catalog.PoolXMLFile import _getPoolCatalogs


def getOutputType(outputs, inputs, directory=""):
    """This function searches the directory for POOL XML catalog files and
    extracts the type of the pfn.

    If not found, inherits from the type of the inputs
    """

    if not isinstance(outputs, list):
        outputs = [outputs]

    catalog = PoolXMLCatalog(_getPoolCatalogs(directory))

    # inputs - by lfn
    generatedIn = False
    typeFileIn = []
    for fname in inputs:
        try:
            tFileIn = str(catalog.getTypeByPfn(str(list(catalog.getPfnsByLfn(fname)["Replicas"].values())[0])))
        except KeyError:
            tFileIn = None
        if not tFileIn:
            generatedIn = True
        else:
            typeFileIn.append(tFileIn)

    if generatedIn and inputs:
        raise ValueError("Could not find Type for inputs")

    # outputs - by pfn
    pfnTypesOut = {}
    for fname in outputs:
        tFileOut = str(catalog.getTypeByPfn(fname))
        if not tFileOut:
            if typeFileIn:
                tFileOut = typeFileIn[0]
            else:
                tFileOut = "ROOT"
        pfnTypesOut[fname] = tFileOut

    return pfnTypesOut



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""This is the LHCb Online storage."""
from xmlrpc import client as xmlrpclib
import os

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Resources.Storage.StorageBase import StorageBase

from DIRAC.Resources.Storage.Utilities import checkArgumentFormat
from DIRAC.Core.Utilities.List import breakListIntoChunks


class LHCbOnlineStorage(StorageBase):
    """Plugin to talk to the xmlrpc of the datamover."""

    def __init__(self, storageName, parameterDict):
        self.isok = True

        super().__init__(storageName, parameterDict)
        self.pluginName = "LHCbOnline"
        self.name = storageName
        self.timeout = 100

        serverString = "{}://{}:{}".format(
            self.protocolParameters["Protocol"],
            self.protocolParameters["Host"],
            self.protocolParameters["Port"],
        )
        self.server = xmlrpclib.Server(serverString)

    def getFileSize(self, urls):
        # FIXME: What the hell is this method doing ??
        """Get a fake file size."""
        if not urls:
            return S_ERROR("LHCbOnline.getFileSize: No surls supplied.")
        successful = {}
        failed = {}
        for pfn in urls:
            successful[pfn] = 0
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def retransferOnlineFile(self, urls):
        """Tell the Online system that the migration failed and we want to get the
        request again."""

        if not urls:
            return S_ERROR("LHCbOnline.requestRetransfer: No surls supplied.")
        successful = {}
        failed = {}
        for pfn in urls:
            try:
                success, error = self.server.errorMigratingFile(pfn)
                if success:
                    successful[pfn] = True
                    gLogger.info("LHCbOnline.requestRetransfer: Successfully requested file from RunDB.")
                else:
                    errStr = f"LHCbOnline.requestRetransfer: Failed to request file from RunDB: {error}"
                    failed[pfn] = errStr
                    gLogger.error(errStr, pfn)
            except Exception as x:  # pylint: disable=broad-except
                errStr = "LHCbOnline.requestRetransfer: Exception while requesting file from RunDB."
                gLogger.exception(errStr, lException=x)
                failed[pfn] = errStr
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def removeFile(self, urls):
        """Remove physically the file specified by its path."""
        if not urls:
            return S_ERROR("LHCbOnline.removeFile: No surls supplied.")
        successful = {}
        failed = {}
        # Here we are sure of the unicity of the basename since it is for raw data only
        filesToUrls = {os.path.basename(f): f for f in urls}
        filenames = list(filesToUrls)

        for filenameChunck in breakListIntoChunks(filenames, 100):
            try:
                success, errorOrFailed = self.server.endMigratingFileBulk(filenameChunck)
                if success:
                    # in case of success, errorOrFailed contains the files for which it failed
                    failedFiles = set(errorOrFailed)
                    for fn in filenameChunck:
                        fullUrl = filesToUrls[fn]
                        if fn in failedFiles:
                            failed[fullUrl] = "Failed to remove, check datamover logs"
                        else:
                            successful[fullUrl] = True
                    gLogger.info("LHCbOnline.getFile: Successfully issued removal to RunDB for chuck.")
                else:
                    errStr = f"LHCbOnline.removeFile: Failed to issue removal for chunck to RunDB {errorOrFailed}"
                    for fn in filenameChunck:
                        fullUrl = filesToUrls[fn]
                        failed[fullUrl] = errStr
                    gLogger.error(errStr, urls)
            except Exception as x:  # pylint: disable=broad-except
                errStr = "LHCbOnline.getFile: Exception for chunck while issuing removal to RunDB."
                gLogger.exception(errStr, lException=x)
                for fn in filenameChunck:
                    fullUrl = filesToUrls[fn]
                    failed[fullUrl] = errStr

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unittest for: LHCbDIRAC.AccountingSystem.private.Plotters.StoragePlotter.

StoragePlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter

We are assuming there is a solid test of __bases__, we are not testing them
here and assuming they work fine.

IMPORTANT: the test MUST be pylint compliant !
"""
# pylint: disable=protected-access

import unittest
from decimal import Decimal

from unittest import mock


class StoragePlotterTestCase(unittest.TestCase):
    """StoragePlotterTestCase."""

    moduleTested = None
    classsTested = None

    def mockModuleTested(self, moduleTested):
        """Used to not redo the mocking done on the parent class ( if any )"""

        # Tries to get the mocks of the parent TestCases ( if any )
        for baseClass in self.__class__.__bases__:
            try:
                # pylint: disable=no-member
                moduleTested = baseClass.mockModuleTested(moduleTested)
            except TypeError:
                continue

        # And then makes its own mock
        class MockStorage:
            # pylint: disable=missing-docstring,no-init
            definitionKeyFields = ("StorageElement", "Directory")

        moduleTested.Storage = mock.Mock(return_value=MockStorage())

        return moduleTested

    def setUp(self):
        """Setup the test case."""

        import LHCbDIRAC.AccountingSystem.private.Plotters.StoragePlotter as moduleTested

        self.moduleTested = self.mockModuleTested(moduleTested)
        self.classsTested = self.moduleTested.StoragePlotter

    def tearDown(self):
        """Tear down the test case."""

        del self.moduleTested
        del self.classsTested


# ...............................................................................


class StoragePlotterUnitTest(StoragePlotterTestCase):
    """DataStoragePlotterUnitTest.

    <constructor>
     - test_instantiate
    <class variables>
     - test_typeName
     - test_typeKeyFields
     - test_reportCatalogSpaceName
     - test_reportCatalogFilesName
     - test_reportPhysicalSpaceName
     - test_reportPhysicalFilesName
     - test_reportPFNvsLFNFileMultiplicityName
     - test_reportPFNvsLFNSizeMultiplicityName
    <methods>
     - test_reportCatalogSpace
     - test_plotCatalogSpace
     - test_reportCatalogFiles
     - test_plotCatalogFiles
     - test_reportPhysicalSpace
     - test_plotPhysicalSpace
     - test_reportPhysicalFiles
     - test_plotPhysicalFiles
    """

    # FIXME: missing test_reportPFNvsLFNFileMultiplicity
    # FIXME: missing test_plotPFNvsLFNFileMultiplicity
    # FIXME: missing test_reportPFNvsLFNSizeMultiplicity
    # FIXME: missing test_plotPFNvsLFNSizeMultiplicity
    # FIXME: missing test_multiplicityReport

    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class."""
        obj = self.classsTested(None, None)
        self.assertEqual("StoragePlotter", obj.__class__.__name__)

    def test_typeName(self):
        """test the class variable "_typeName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._typeName, "Storage")

    def test_typeKeyFields(self):
        """test the class variable "_typeKeyFields"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._typeKeyFields, ["StorageElement", "Directory"])

    def test_reportCatalogSpaceName(self):
        """test the class variable "_reportCatalogSpaceName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportCatalogSpaceName, "LFN size")

    def test_reportCatalogFilesName(self):
        """test the class variable "_reportCatalogFilesName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportCatalogFilesName, "LFN files")

    def test_reportPhysicalSpaceName(self):
        """test the class variable "_reportPhysicalSpaceName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportPhysicalSpaceName, "PFN size")

    def test_reportPhysicalFilesName(self):
        """test the class variable "_reportPhysicalFilesName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportPhysicalFilesName, "PFN files")

    def test_reportPFNvsLFNFileMultiplicityName(self):
        """test the class variable "_reportPFNvsLFNFileMultiplicityName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportPFNvsLFNFileMultiplicityName, "PFN/LFN file ratio")

    def test_reportPFNvsLFNSizeMultiplicityName(self):
        """test the class variable "_reportPFNvsLFNSizeMultiplicityName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportPFNvsLFNSizeMultiplicityName, "PFN/LFN size ratio")

    def test_reportCatalogSpace(self):
        """test the method "_reportCatalogSpace"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        res = obj._reportCatalogSpace({"grouping": "StorageElement"})
        self.assertEqual(res["OK"], False, msg="Rejected StorageElement grouping")
        self.assertEqual(res["Message"], "Grouping by storage element when requesting lfn info makes no sense")

        #     res = obj._reportCatalogSpace( { 'grouping'       : 'NextToABeer',
        #                                      'groupingFields' : [ 0, [ 'mehh' ], 'blah' ],
        #                                      'startTime'      : 'startTime',
        #                                      'endTime'        : 'endTime',
        #                                      'condDict'       : {}
        #                                     } )
        #     self.assertEqual( res[ 'OK' ], False )
        #     self.assertEqual( res[ 'Message' ], 'No connection' )

        # Changed mocked to run over different lines of code
        mockAccountingDB._getConnection.return_value = {"OK": True, "Value": []}
        res = obj._reportCatalogSpace(
            {
                "grouping": "NextToABeer",
                "groupingFields": [0, ["mehh"], "blah"],
                "startTime": "startTime",
                "endTime": "endTime",
                "condDict": {},
            }
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(res["Value"], {"graphDataDict": {}, "data": {}, "unit": "MB", "granularity": "BucketLength"})

        mockedData = (
            ("/lhcb/data", 1355616000, 86400, Decimal("4935388524246.91")),
            ("/lhcb/data", 1355702400, 86400, Decimal("4843844487074.82")),
            ("/lhcb/LHCb", 1355616000, 86400, Decimal("3935388524246.91")),
            ("/lhcb/LHCb", 1355702400, 86400, Decimal("3843844487074.82")),
        )
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": mockedData}
        mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

        res = obj._reportCatalogSpace(
            {
                "grouping": "Directory",
                "groupingFields": ("%s", ["Directory"]),
                "startTime": 1355663249.0,
                "endTime": 1355749690.0,
                "condDict": {"Directory": ["/lhcb/data", "/lhcb/LHCb"]},
            }
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(
            res["Value"],
            {
                "graphDataDict": {
                    "/lhcb/data": {1355616000: 4.9353885242469104, 1355702400: 4.8438444870748203},
                    "/lhcb/LHCb": {1355616000: 3.93538852424691, 1355702400: 3.8438444870748198},
                },
                "data": {
                    "/lhcb/data": {1355616000: 4935388.5242469106, 1355702400: 4843844.4870748203},
                    "/lhcb/LHCb": {1355616000: 3935388.5242469101, 1355702400: 3843844.4870748199},
                },
                "unit": "TB",
                "granularity": 86400,
            },
        )

    def test_reportCatalogFiles(self):
        """test the method "_reportCatalogFiles"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        res = obj._reportCatalogFiles({"grouping": "StorageElement"})
        self.assertEqual(res["OK"], False)
        self.assertEqual(res["Message"], "Grouping by storage element when requesting lfn info makes no sense")

        #     res = obj._reportCatalogFiles( { 'grouping'       : 'NextToABeer',
        #                                      'groupingFields' : [ 0, [ 'mehh' ], 'blah' ],
        #                                      'startTime'      : 'startTime',
        #                                      'endTime'        : 'endTime',
        #                                      'condDict'       : {}
        #                                     } )
        #     self.assertEqual( res[ 'OK' ], False )
        #     self.assertEqual( res[ 'Message' ], 'No connection' )

        # Changed mocked to run over different lines of code
        mockAccountingDB._getConnection.return_value = {"OK": True, "Value": []}
        res = obj._reportCatalogFiles(
            {
                "grouping": "NextToABeer",
                "groupingFields": [0, ["mehh"], "blah"],
                "startTime": "startTime",
                "endTime": "endTime",
                "condDict": {},
            }
        )
        self.assertEqual(res["OK"], True, msg="Expected S_OK")
        self.assertEqual(
            res["Value"], {"graphDataDict": {}, "data": {}, "unit": "files", "granularity": "BucketLength"}
        )

        mockedData = (
            ("/lhcb/data", 1355616000, 86400, Decimal("4935388524246.91")),
            ("/lhcb/data", 1355702400, 86400, Decimal("4843844487074.82")),
            ("/lhcb/LHCb", 1355616000, 86400, Decimal("3935388524246.91")),
            ("/lhcb/LHCb", 1355702400, 86400, Decimal("3843844487074.82")),
        )
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": mockedData}
        mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

        res = obj._reportCatalogFiles(
            {
                "grouping": "Directory",
                "groupingFields": ("%s", ["Directory"]),
                "startTime": 1355663249.0,
                "endTime": 1355749690.0,
                "condDict": {"Directory": ["/lhcb/data", "/lhcb/LHCb"]},
            }
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(
            res["Value"],
            {
                "graphDataDict": {
                    "/lhcb/data": {1355616000: 4935388.5242469106, 1355702400: 4843844.4870748203},
                    "/lhcb/LHCb": {1355616000: 3935388.5242469101, 1355702400: 3843844.4870748199},
                },
                "data": {
                    "/lhcb/data": {1355616000: 4935388524246.9102, 1355702400: 4843844487074.8203},
                    "/lhcb/LHCb": {1355616000: 3935388524246.9102, 1355702400: 3843844487074.8198},
                },
                "unit": "Mfiles",
                "granularity": 86400,
            },
        )

    def test_reportPhysicalSpace(self):
        """test the method "_reportPhysicalSpace"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        #     res = obj._reportPhysicalSpace( { 'groupingFields' : [ 0, [ 'mehh' ], 'blah' ],
        #                                       'startTime'      : 'startTime',
        #                                       'endTime'        : 'endTime',
        #                                       'condDict'       : {}
        #                                      } )
        #     self.assertEqual( res[ 'OK' ], False )
        #     self.assertEqual( res[ 'Message' ], 'No connection' )

        # Changed mocked to run over different lines of code
        mockAccountingDB._getConnection.return_value = {"OK": True, "Value": []}
        res = obj._reportPhysicalSpace(
            {
                "grouping": "NextToABeer",
                "groupingFields": [0, ["mehh"], "blah"],
                "startTime": "startTime",
                "endTime": "endTime",
                "condDict": {},
            }
        )
        self.assertEqual(res["OK"], True, msg="Expected S_OK")
        self.assertEqual(res["Value"], {"graphDataDict": {}, "data": {}, "unit": "MB", "granularity": "BucketLength"})

        mockedData = (
            ("CERN-ARCHIVE", 1355616000, 86400, Decimal("2344556767812.91")),
            ("CERN-ARCHIVE", 1355702400, 86400, Decimal("2544556767812.91")),
            ("CERN-DST", 1355616000, 86400, Decimal("344556767812.91")),
            ("CERN-DST", 1355702400, 86400, Decimal("544556767812.91")),
        )
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": mockedData}
        mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

        res = obj._reportPhysicalSpace(
            {
                "grouping": "StorageElement",
                "groupingFields": ("%s", ["StorageElement"]),
                "startTime": 1355663249.0,
                "endTime": 1355749690.0,
                "condDict": {"StorageElement": ["CERN-ARCHIVE", "CERN-DST"]},
            }
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(
            res["Value"],
            {
                "graphDataDict": {
                    "CERN-ARCHIVE": {1355616000: 2.34455676781291, 1355702400: 2.5445567678129102},
                    "CERN-DST": {1355616000: 0.34455676781290995, 1355702400: 0.54455676781290996},
                },
                "data": {
                    "CERN-ARCHIVE": {1355616000: 2344556.76781291, 1355702400: 2544556.76781291},
                    "CERN-DST": {1355616000: 344556.76781290997, 1355702400: 544556.76781291002},
                },
                "unit": "TB",
                "granularity": 86400,
            },
        )

    def test_reportPhysicalFiles(self):
        """test the method "_reportPhysicalFiles"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        #     res = obj._reportPhysicalFiles( { 'groupingFields' : [ 0, [ 'mehh' ], 'blah' ],
        #                                       'startTime'      : 'startTime',
        #                                       'endTime'        : 'endTime',
        #                                       'condDict'       : {}
        #                                      } )
        #     self.assertEqual( res[ 'OK' ], False )
        #     self.assertEqual( res[ 'Message' ], 'No connection' )

        # Changed mocked to run over different lines of code
        mockAccountingDB._getConnection.return_value = {"OK": True, "Value": []}
        res = obj._reportPhysicalFiles(
            {
                "grouping": "NextToABeer",
                "groupingFields": [0, ["mehh"], "blah"],
                "startTime": "startTime",
                "endTime": "endTime",
                "condDict": {},
            }
        )
        self.assertEqual(res["OK"], True, msg="Expected S_OK")
        self.assertEqual(
            res["Value"], {"graphDataDict": {}, "data": {}, "unit": "files", "granularity": "BucketLength"}
        )

        mockedData = (
            ("CERN-ARCHIVE", 1355616000, 86400, Decimal("412658.91")),
            ("CERN-ARCHIVE", 1355702400, 86400, Decimal("413658.91")),
            ("CERN-BUFFER", 1355616000, 86400, Decimal("250658.91")),
            ("CERN-BUFFER", 1355702400, 86400, Decimal("261658.91")),
            ("CERN-DST", 1355616000, 86400, Decimal("186658.91")),
            ("CERN-DST", 1355702400, 86400, Decimal("187658.91")),
        )
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": mockedData}
        mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

        res = obj._reportPhysicalFiles(
            {
                "grouping": "StorageElement",
                "groupingFields": ("%s", ["StorageElement"]),
                "startTime": 1355663249.0,
                "endTime": 1355749690.0,
                "condDict": {"StorageElement": ["CERN-ARCHIVE", "CERN-DST", "CERN-BUFFER"]},
            }
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(
            res["Value"],
            {
                "graphDataDict": {
                    "CERN-BUFFER": {1355616000: 250.65890999999999, 1355702400: 261.65890999999999},
                    "CERN-ARCHIVE": {1355616000: 412.65890999999999, 1355702400: 413.65890999999999},
                    "CERN-DST": {1355616000: 186.65890999999999, 1355702400: 187.65890999999999},
                },
                "data": {
                    "CERN-BUFFER": {1355616000: 250658.91, 1355702400: 261658.91},
                    "CERN-ARCHIVE": {1355616000: 412658.90999999997, 1355702400: 413658.90999999997},
                    "CERN-DST": {1355616000: 186658.91, 1355702400: 187658.91},
                },
                "unit": "kfiles",
                "granularity": 86400,
            },
        )

    def test_multiplicityReport(self):
        """test the method "_multiplicityReport"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        #     res = obj._multiplicityReport( { 'groupingFields' : [ 0, [ 'mehh' ], 'blah' ],
        #                                      'startTime'      : 'startTime',
        #                                      'endTime'        : 'endTime',
        #                                      'condDict'       : {}
        #                                      }, '', '' )
        #     self.assertEqual( res[ 'OK' ], False )
        #     self.assertEqual( res[ 'Message' ], 'No connection' )
        #
        # We are using mock 0.7, we need this little hack
        side_values = [{"OK": False, "Message": "No connection"}, {"OK": True, "Value": []}]

        def side_effect():
            return side_values.pop()

        # Changed mocked to run over different lines of code
        #     mockAccountingDB._getConnection.side_effect = side_effect
        #     res = obj._multiplicityReport( { 'groupingFields' : [ 0, [ 'mehh' ], 'blah' ],
        #                                      'startTime'      : 'startTime',
        #                                      'endTime'        : 'endTime',
        #                                      'condDict'       : {}
        #                                     }, '', '' )
        #     self.assertEqual( res[ 'OK' ], False )
        #     self.assertEqual( res[ 'Message' ], 'No connection' )

        # Changed mocked to run over different lines of code
        mockAccountingDB._getConnection.return_value = {"OK": True, "Value": []}


# ...............................................................................


class StoragePlotterUnitTestCrashes(StoragePlotterTestCase):
    """StoragePlotterUnitTestCrashes.

    <constructor>
     - test_instantiate
    <class variables>
    <methods>
     - test_reportCatalogSpace
     - test_plotCatalogSpace
     - test_reportCatalogFiles
     - test_plotCatalogFiles
     - test_reportPhysicalSpace
     - test_plotPhysicalSpace
     - test_reportPhysicalFiles
     - test_plotPhysicalFiles
    """

    def test_instantiate(self):
        """test the constructor."""

        self.assertRaises(TypeError, self.classsTested)
        self.assertRaises(TypeError, self.classsTested, None)
        self.assertRaises(TypeError, self.classsTested, None, None, None, None)

        self.assertRaises(TypeError, self.classsTested, extraArgs=None)
        self.assertRaises(TypeError, self.classsTested, None, extraArgs=None)
        self.assertRaises(TypeError, self.classsTested, None, None, None, extraArgs=None)

    def test_reportCatalogSpace(self):
        """test the method "_reportCatalogSpace"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        self.assertRaises(KeyError, obj._reportCatalogSpace, {})
        self.assertRaises(KeyError, obj._reportCatalogSpace, {"grouping": 1})
        self.assertRaises(IndexError, obj._reportCatalogSpace, {"grouping": 1, "groupingFields": []})
        self.assertRaises(TypeError, obj._reportCatalogSpace, {"grouping": 1, "groupingFields": [1, 2]})
        self.assertRaises(TypeError, obj._reportCatalogSpace, {"grouping": 1, "groupingFields": [1, [2]]})
        self.assertRaises(TypeError, obj._reportCatalogSpace, {"grouping": 1, "groupingFields": ["1", "2"]})
        self.assertRaises(KeyError, obj._reportCatalogSpace, {"grouping": 1, "groupingFields": ["1", [2]]})
        self.assertRaises(
            KeyError, obj._reportCatalogSpace, {"grouping": 1, "groupingFields": ["1", [2, 2]], "startTime": None}
        )
        self.assertRaises(
            KeyError,
            obj._reportCatalogSpace,
            {"grouping": 1, "groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None},
        )
        self.assertRaises(
            TypeError,
            obj._reportCatalogSpace,
            {"grouping": 1, "groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None, "condDict": None},
        )

    def test_plotCatalogSpace(self):
        """test the method "_plotCatalogSpace"."""

        obj = self.classsTested(None, None)
        self.assertRaises(TypeError, obj._plotCatalogSpace, None, None, None)
        self.assertRaises(KeyError, obj._plotCatalogSpace, {}, None, None)
        self.assertRaises(KeyError, obj._plotCatalogSpace, {"startTime": "startTime"}, None, None)
        self.assertRaises(
            TypeError, obj._plotCatalogSpace, {"startTime": "startTime", "endTime": "endTime"}, None, None
        )
        self.assertRaises(KeyError, obj._plotCatalogSpace, {"startTime": "startTime", "endTime": "endTime"}, {}, None)
        self.assertRaises(
            KeyError,
            obj._plotCatalogSpace,
            {"startTime": "startTime", "endTime": "endTime"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotCatalogSpace,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotCatalogSpace,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity", "graphDataDict": "graphDataDict"},
            None,
        )

    def test_reportCatalogFiles(self):
        """test the method "_reportCatalogFiles"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        self.assertRaises(KeyError, obj._reportCatalogFiles, {})
        self.assertRaises(KeyError, obj._reportCatalogFiles, {"grouping": 1})
        self.assertRaises(IndexError, obj._reportCatalogFiles, {"grouping": 1, "groupingFields": []})
        self.assertRaises(TypeError, obj._reportCatalogFiles, {"grouping": 1, "groupingFields": [1, 2]})
        self.assertRaises(TypeError, obj._reportCatalogFiles, {"grouping": 1, "groupingFields": [1, [2]]})
        self.assertRaises(TypeError, obj._reportCatalogFiles, {"grouping": 1, "groupingFields": ["1", "2"]})
        self.assertRaises(KeyError, obj._reportCatalogFiles, {"grouping": 1, "groupingFields": ["1", [2]]})
        self.assertRaises(
            KeyError, obj._reportCatalogFiles, {"grouping": 1, "groupingFields": ["1", [2, 2]], "startTime": None}
        )
        self.assertRaises(
            KeyError,
            obj._reportCatalogFiles,
            {"grouping": 1, "groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None},
        )
        self.assertRaises(
            TypeError,
            obj._reportCatalogFiles,
            {"grouping": 1, "groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None, "condDict": None},
        )

    def test_plotCatalogFiles(self):
        """test the method "_plotCatalogFiles"."""

        obj = self.classsTested(None, None)
        self.assertRaises(TypeError, obj._plotCatalogFiles, None, None, None)
        self.assertRaises(KeyError, obj._plotCatalogFiles, {}, None, None)
        self.assertRaises(KeyError, obj._plotCatalogFiles, {"startTime": "startTime"}, None, None)
        self.assertRaises(
            TypeError, obj._plotCatalogFiles, {"startTime": "startTime", "endTime": "endTime"}, None, None
        )
        self.assertRaises(KeyError, obj._plotCatalogFiles, {"startTime": "startTime", "endTime": "endTime"}, {}, None)
        self.assertRaises(
            KeyError,
            obj._plotCatalogFiles,
            {"startTime": "startTime", "endTime": "endTime"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotCatalogFiles,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotCatalogFiles,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity", "graphDataDict": "graphDataDict"},
            None,
        )

    def test_reportPhysicalSpace(self):
        """test the method "_reportPhysicalSpace"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        self.assertRaises(KeyError, obj._reportPhysicalSpace, {})
        self.assertRaises(IndexError, obj._reportPhysicalSpace, {"groupingFields": []})
        self.assertRaises(TypeError, obj._reportPhysicalSpace, {"groupingFields": [1, 2]})
        self.assertRaises(TypeError, obj._reportPhysicalSpace, {"groupingFields": [1, [2]]})
        self.assertRaises(TypeError, obj._reportPhysicalSpace, {"groupingFields": ["1", "2"]})
        self.assertRaises(KeyError, obj._reportPhysicalSpace, {"groupingFields": ["1", [2]]})
        self.assertRaises(KeyError, obj._reportPhysicalSpace, {"groupingFields": ["1", [2, 2]], "startTime": None})
        self.assertRaises(
            KeyError, obj._reportPhysicalSpace, {"groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None}
        )
        self.assertRaises(
            TypeError,
            obj._reportPhysicalSpace,
            {"groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None, "condDict": None},
        )

    def test_plotPhysicalSpace(self):
        """test the method "_plotPhysicalSpace"."""

        obj = self.classsTested(None, None)
        self.assertRaises(TypeError, obj._plotPhysicalSpace, None, None, None)
        self.assertRaises(KeyError, obj._plotPhysicalSpace, {}, None, None)
        self.assertRaises(KeyError, obj._plotPhysicalSpace, {"startTime": "startTime"}, None, None)
        self.assertRaises(
            TypeError, obj._plotPhysicalSpace, {"startTime": "startTime", "endTime": "endTime"}, None, None
        )
        self.assertRaises(KeyError, obj._plotPhysicalSpace, {"startTime": "startTime", "endTime": "endTime"}, {}, None)
        self.assertRaises(
            KeyError,
            obj._plotPhysicalSpace,
            {"startTime": "startTime", "endTime": "endTime"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotPhysicalSpace,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotPhysicalSpace,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity", "graphDataDict": "graphDataDict"},
            None,
        )

    def test_reportPhysicalFiles(self):
        """test the method "_reportPhysicalFiles"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        self.assertRaises(KeyError, obj._reportPhysicalFiles, {})
        self.assertRaises(IndexError, obj._reportPhysicalFiles, {"groupingFields": []})
        self.assertRaises(TypeError, obj._reportPhysicalFiles, {"groupingFields": [1, 2]})
        self.assertRaises(TypeError, obj._reportPhysicalFiles, {"groupingFields": [1, [2]]})
        self.assertRaises(TypeError, obj._reportPhysicalFiles, {"groupingFields": ["1", "2"]})
        self.assertRaises(KeyError, obj._reportPhysicalFiles, {"groupingFields": ["1", [2]]})
        self.assertRaises(KeyError, obj._reportPhysicalFiles, {"groupingFields": ["1", [2, 2]], "startTime": None})
        self.assertRaises(
            KeyError, obj._reportPhysicalFiles, {"groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None}
        )
        self.assertRaises(
            TypeError,
            obj._reportPhysicalFiles,
            {"groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None, "condDict": None},
        )

    def test_plotPhysicalFiles(self):
        """test the method "_plotPhysicalFiles"."""

        obj = self.classsTested(None, None)
        self.assertRaises(TypeError, obj._plotPhysicalFiles, None, None, None)
        self.assertRaises(KeyError, obj._plotPhysicalFiles, {}, None, None)
        self.assertRaises(KeyError, obj._plotPhysicalFiles, {"startTime": "startTime"}, None, None)
        self.assertRaises(
            TypeError, obj._plotPhysicalFiles, {"startTime": "startTime", "endTime": "endTime"}, None, None
        )
        self.assertRaises(KeyError, obj._plotPhysicalFiles, {"startTime": "startTime", "endTime": "endTime"}, {}, None)
        self.assertRaises(
            KeyError,
            obj._plotPhysicalFiles,
            {"startTime": "startTime", "endTime": "endTime"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotPhysicalFiles,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotPhysicalFiles,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity", "graphDataDict": "graphDataDict"},
            None,
        )


################################################################################


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(StoragePlotterTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(StoragePlotterUnitTest))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(StoragePlotterUnitTestCrashes))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""stores the replica readed from an xml."""
from DIRAC import gLogger


class Replica:
    """Replica class."""

    #############################################################################
    def __init__(self):
        """initialize the class members."""
        self.params = []
        self.name = ""

    #############################################################################
    def addParam(self, param):
        """sets the parameters."""
        self.params += [param]

    #############################################################################
    def __repr__(self):
        """It idents the print output."""
        result = "\nReplica: "
        result += self.name + "\n"
        for param in self.params:
            result += str(param)

        return result

    #############################################################################
    def writeToXML(self):
        """writs an XML file."""
        gLogger.debug("Replica XML writing!!!")
        result = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE Replicas SYSTEM "book.dtd">
<Replicas>
"""
        for param in self.params:
            result += param.writeToXML(False)

        result += "</Replicas>"
        return result

    #############################################################################



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""stores the data quality informations."""


class QualityParameters:
    def __init__(self):
        """initialize the class members."""
        self.name = ""
        self.value = ""

    def __repr__(self):
        """formats the output of the print command."""
        return self.name + " " + self.value + "\n"

    def writeToXML(self):
        """creates an xml string."""
        return '  <Parameter Name="' + str(self.name) + '" Value="' + str(self.value) + '"/>\n'



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC Job DB.

Extends the DIRAC JobDB with minor things
"""
from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB as DIRACJobDB


class JobDB(DIRACJobDB):
    """Extension of the DIRAC Job DB."""

    def __init__(self, *args, **kwargs):
        """The standard constructor takes the database name (dbname) and the name
        of the configuration section (dbconfig)"""
        DIRACJobDB.__init__(self, *args, **kwargs)
        self.jdl2DBParameters += ["runNumber"]



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" WMSSecureGW service -  a generic gateway service

    Mostly used by BOINC
"""
import json
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Security import Properties
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getDNForUsername, findDefaultGroupForDN
from DIRAC.AccountingSystem.Client.DataStoreClient import DataStoreClient
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers import cfgPath
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.WorkloadManagementSystem.Client.MatcherClient import MatcherClient
from DIRAC.WorkloadManagementSystem.Client.JobStateUpdateClient import JobStateUpdateClient
from DIRAC.WorkloadManagementSystem.Client.JobManagerClient import JobManagerClient
from DIRAC.WorkloadManagementSystem.Client.PilotManagerClient import PilotManagerClient
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient

# pylint: disable=no-self-use


class WMSSecureGWHandler(RequestHandler):
    """WMSSecure class."""

    @classmethod
    def initializeHandler(cls, serviceInfo):  # pylint: disable=unused-argument
        """Handler initialization."""
        from DIRAC.DataManagementSystem.Service.FileCatalogHandler import FileCatalogHandler

        if FileCatalogHandler.types_hasAccess != cls.types_hasAccess:
            raise Exception("FileCatalog hasAccess types has been changed.")
        if FileCatalogHandler.types_exists != cls.types_exists:
            raise Exception("FileCatalog exists types has been changed.")
        if FileCatalogHandler.types_addFile != cls.types_addFile:
            raise Exception("FileCatalog addFile types has been changed.")
        from DIRAC.WorkloadManagementSystem.Service.MatcherHandler import MatcherHandler

        if MatcherHandler.types_requestJob != cls.types_requestJob:
            raise Exception("Matcher requestJob types has been changed.")
        from DIRAC.WorkloadManagementSystem.Service.JobStateUpdateHandler import JobStateUpdateHandler

        if JobStateUpdateHandler.types_setJobStatus != cls.types_setJobStatus:
            raise Exception("JobStateUpdate setJobStatus types has been changed.")
        if JobStateUpdateHandler.types_setJobSite != cls.types_setJobSite:
            raise Exception("JobStateUpdate setJobSite types has been changed.")
        if JobStateUpdateHandler.types_setJobParameter != cls.types_setJobParameter:
            raise Exception("JobStateUpdate setJobParameter types has been changed.")
        if JobStateUpdateHandler.types_setJobStatusBulk != cls.types_setJobStatusBulk:
            raise Exception("JobStateUpdate setJobStatusBulk types has been changed.")
        if JobStateUpdateHandler.types_setJobParameters != cls.types_setJobParameters:
            raise Exception("JobStateUpdate setJobParameters types has been changed.")
        if JobStateUpdateHandler.types_sendHeartBeat != cls.types_sendHeartBeat:
            raise Exception("JobStateUpdate sendHeartBeat types has been changed.")
        from DIRAC.WorkloadManagementSystem.Service.JobManagerHandler import JobManagerHandler

        if JobManagerHandler.types_rescheduleJob != cls.types_rescheduleJob:
            raise Exception("JobManager rescheduleJob types has been changed.")
        from DIRAC.WorkloadManagementSystem.Service.PilotManagerHandler import PilotManagerHandler

        if PilotManagerHandler.types_setPilotStatus != cls.types_setPilotStatus:
            raise Exception("PilotManager setPilotStatus types has been changed.")
        if PilotManagerHandler.types_setJobForPilot != cls.types_setJobForPilot:
            raise Exception("PilotManager setJobForPilot types has been changed.")
        if PilotManagerHandler.types_setPilotBenchmark != cls.types_setPilotBenchmark:
            raise Exception("PilotManager setPilotBenchmark types has been changed.")
        from DIRAC.WorkloadManagementSystem.Service.JobMonitoringHandler import JobMonitoringHandler

        if JobMonitoringHandler.types_getJobParameter != cls.types_getJobParameter:
            raise Exception("JobMonitoring getJobParameter types has been changed.")
        from DIRAC.FrameworkSystem.Service.ProxyManagerHandler import ProxyManagerHandler

        if ProxyManagerHandler.types_getVOMSProxy != cls.types_getVOMSProxy:
            raise Exception("ProxyManagerHandler getVOMSProxy types has been changed.")
        if ProxyManagerHandler.types_getProxy != cls.types_getProxy:
            raise Exception("ProxyManagerHandler getProxy types has been changed.")
        from DIRAC.RequestManagementSystem.Service.ReqManagerHandler import ReqManagerHandler

        if ReqManagerHandler.types_putRequest != cls.types_putRequest:
            raise Exception("ReqManagerHandler putRequest types has been changed.")
        from DIRAC.AccountingSystem.Service.DataStoreHandler import DataStoreHandler

        if DataStoreHandler.types_commitRegisters != cls.types_commitRegisters:
            raise Exception("DataStoreHandler commitRegisters types has been changed.")

        return S_OK()

    types_requestJob = [[(str,), dict]]

    def export_requestJob(self, resourceDescription):
        """Serve a job to the request of an agent which is the highest priority one
        matching the agent's site capacity."""
        result = MatcherClient(timeout=600).requestJob(resourceDescription)
        return result

    ###########################################################################
    types_setJobStatus = [((str,), (int,)), (str,), (str,), (str,)]

    def export_setJobStatus(self, jobID, status, minorStatus, source="Unknown", datetime=None):
        """Set the major and minor status for job specified by its JobId.

        Set optionally the status date and source component which sends the
        status information.
        """
        jobStatus = JobStateUpdateClient().setJobStatus(int(jobID), status, minorStatus, source, datetime)
        return jobStatus

    ###########################################################################
    types_setJobSite = [((str,), (int,)), (str,)]

    def export_setJobSite(self, jobID, site):
        """Allows the site attribute to be set for a job specified by its jobID."""
        jobSite = JobStateUpdateClient().setJobSite(jobID, site)
        return jobSite

    ###########################################################################
    types_setJobParameter = [((str,), (int,)), (str,), (str,)]

    def export_setJobParameter(self, jobID, name, value):
        """Set arbitrary parameter specified by name/value pair for job specified
        by its JobId."""
        jobParam = JobStateUpdateClient().setJobParameter(int(jobID), name, value)
        return jobParam

    ###########################################################################
    types_setJobStatusBulk = [((str,), (int,)), dict]

    def export_setJobStatusBulk(self, jobID, statusDict):
        """Set various status fields for job specified by its JobId.

        Set only the last status in the JobDB, updating all the status
        logging information in the JobLoggingDB. The statusDict has datetime
        as a key and status information dictionary as values
        """
        jobStatus = JobStateUpdateClient().setJobStatusBulk(jobID, statusDict)
        return jobStatus

    ###########################################################################
    types_setJobParameters = [((str,), (int,)), list]

    def export_setJobParameters(self, jobID, parameters):
        """Set arbitrary parameters specified by a list of name/value pairs for job
        specified by its JobId."""
        jobParams = JobStateUpdateClient().setJobParameters(jobID, parameters)
        return jobParams

    ###########################################################################
    types_sendHeartBeat = [((str,), (int,)), dict, dict]

    def export_sendHeartBeat(self, jobID, dynamicData, staticData):
        """Send a heart beat sign of life for a job jobID."""
        result = JobStateUpdateClient(timeout=120).sendHeartBeat(jobID, dynamicData, staticData)
        return result

    ##########################################################################################
    types_rescheduleJob = []

    def export_rescheduleJob(self, jobIDs):
        """Reschedule a single job.

        If the optional proxy parameter is given it will be used to refresh
        the proxy in the Proxy Repository
        """
        result = JobManagerClient().rescheduleJob(jobIDs)
        return result

    ##########################################################################################
    types_setPilotStatus = [(str,), (str,)]

    def export_setPilotStatus(self, pilotRef, status, destination=None, reason=None, gridSite=None, queue=None):
        """Set the pilot agent status."""
        result = PilotManagerClient().setPilotStatus(pilotRef, status, destination, reason, gridSite, queue)
        return result

    ##############################################################################
    types_setJobForPilot = [(str,) + (int,), (str,)]

    def export_setJobForPilot(self, jobID, pilotRef, destination=None):
        """Report the DIRAC job ID which is executed by the given pilot job."""
        result = PilotManagerClient().setJobForPilot(jobID, pilotRef, destination)
        return result

    ##########################################################################################
    types_setPilotBenchmark = [(str,), float]

    def export_setPilotBenchmark(self, pilotRef, mark):
        """Set the pilot agent benchmark."""
        result = PilotManagerClient().setPilotBenchmark(pilotRef, mark)
        return result

    ##############################################################################
    types_getJobParameter = [((str,), (int,)), (str,)]

    @staticmethod
    def export_getJobParameter(jobID, parName):
        result = JobMonitoringClient(timeout=120).getJobParameter(jobID, parName)
        return result

    ##############################################################################
    types_getVOMSProxy = [(str,), (str,), (str,), (int,)]

    def export_getVOMSProxy(
        self, userDN, userGroup, requestPem, requiredLifetime, vomsAttribute=False
    ):  # pylint: disable=unused-argument
        """Always return the Boinc proxy."""
        userDN, userGroup, _ = self.__getOwnerGroupDN("BoincUser")
        rpcClient = RPCClient("Framework/BoincProxyManager", timeout=120)
        retVal = rpcClient.getProxy(userDN, userGroup, requestPem, requiredLifetime)
        return retVal

    ##############################################################################
    types_getProxy = [(str,), (str,), (str,), (int,)]

    def export_getProxy(self, userDN, userGroup, requestPem, requiredLifetime):  # pylint: disable=unused-argument
        """Get the Boinc User proxy."""
        userDN, userGroup, _ = self.__getOwnerGroupDN("BoincUser")
        rpcClient = RPCClient("Framework/BoincProxyManager", timeout=120)
        retVal = rpcClient.getProxy(userDN, userGroup, requestPem, requiredLifetime)
        return retVal

    ##############################################################################
    def __checkProperties(self, requestedUserDN, requestedUserGroup):
        """Check the properties and return if they can only download limited
        proxies if authorized."""
        credDict = self.getRemoteCredentials()
        gLogger.debug(f"in credDict {credDict['properties']}")
        if Properties.FULL_DELEGATION in credDict["properties"]:
            return S_OK(False)
        if Properties.LIMITED_DELEGATION in credDict["properties"]:
            return S_OK(True)
        if Properties.PRIVATE_LIMITED_DELEGATION in credDict["properties"]:
            if credDict["DN"] != requestedUserDN:
                return S_ERROR("You are not allowed to download any proxy")
            if Properties.PRIVATE_LIMITED_DELEGATION in Registry.getPropertiesForGroup(requestedUserGroup):
                return S_ERROR("You can't download proxies for that group")
            return S_OK(True)
        # Not authorized!
        return S_ERROR("You can't get proxies! Bad boy!")

    ########################################################################

    types_hasAccess = [[(str,), dict], [(str,), list, dict]]

    def export_hasAccess(self, paths, opType):  # pylint: disable=unused-argument
        """Access."""
        successful = {}
        for path in paths:
            successful[path] = True
        resDict = {"Successful": successful, "Failed": {}}
        return S_OK(resDict)

    types_exists = [[list, dict] + list((str,))]

    def export_exists(self, lfns):
        """Check whether the supplied paths exists."""
        successful = {}
        for lfn in lfns:
            successful[lfn] = False
        resDict = {"Successful": successful, "Failed": {}}
        return S_OK(resDict)

    ########################################################################

    types_addFile = [[list, dict] + list((str,))]

    def export_addFile(self, lfns):
        """Register supplied files."""
        failed = {}
        for lfn in lfns:
            failed[lfn] = True
        return S_OK({"Successful": {}, "Failed": failed})

    types_putRequest = [(str,)]

    def export_putRequest(self, requestJSON):
        """put a new request into RequestDB."""

        requestDict = json.loads(requestJSON)
        request = Request(requestDict)
        operation = Operation()  # # create new operation
        operation.Type = "WMSSecureOutputData"
        request.insertBefore(operation, request[0])
        userDN, userGroup, _ = self.__getOwnerGroupDN("ProductionManager")
        request.OwnerDN = userDN
        request.OwnerGroup = userGroup
        return ReqClient().putRequest(request)

    def __getOwnerGroupDN(self, shifterType):
        opsHelper = Operations()
        userName = opsHelper.getValue(cfgPath("BoincShifter", shifterType, "User"), "")
        if not userName:
            return S_ERROR(f"No shifter User defined for {shifterType}")
        result = getDNForUsername(userName)
        if not result["OK"]:
            return result
        userDN = result["Value"][0]
        result = findDefaultGroupForDN(userDN)
        if not result["OK"]:
            return result
        defaultGroup = result["Value"]
        userGroup = opsHelper.getValue(cfgPath("BoincShifter", shifterType, "Group"), defaultGroup)
        return userDN, userGroup, userName

    ########################################################################

    types_commitRegisters = [list]

    def export_commitRegisters(self, entriesList):
        retVal = DataStoreClient().commitRegisters(entriesList)
        return retVal



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""dirac-production-shifter.

Script for the production shifter.

Print a summary of the requests in the system ( and their productions ).
The following switches and options are provided.

- requestState  : request state(s)
   [ Accepted, Active, BK Check, BK OK, Cancelled, Done, New, PPG OK, Rejected,
     Submitted, Tech OK ]
- requestType   : request type(s)
   [ Reconstruction, Simulation, Stripping, Stripping (Moore), Swimming, WGProduction ]
- requestID     : request ID(s)
- simCondition  : simulation condition(s) ( e.g. Beam4000GeV-VeloClosed-MagDown )
- proPath       : proPath(s) ( e.g. Reco13 )
- eventType     : eventType(s) ( e.g. 90000000 )
- sort          : sort requests using the keywords [RequestID,RequestState,RequestType,
                  SimCondition,ProPath,EventType]
- groupMerge    : group all merge productions in one line
- omitMerge     : omit all merge production
- noFiles       : do not request file information
"""

from datetime import datetime
import sys

from DIRAC import exit as DIRACExit
from DIRAC.Core.Base.Script import Script


def doParse():
    """Function that contains all the switches definition and isolates the rest
    of the module from parseCommandLine."""

    # Switch description
    Script.registerSwitch("i:", "requestID=", "ID of the request")
    Script.registerSwitch("r:", "requestState=", "request state(s)")
    Script.registerSwitch("t:", "requestType=", "request type(s)")
    Script.registerSwitch("m:", "simCondition=", "simulation condition(s)")
    Script.registerSwitch("p:", "proPath=", "proPath(s)")
    Script.registerSwitch("e:", "eventType=", "eventType(s)")
    Script.registerSwitch("z:", "sort=", "sort requests")
    Script.registerSwitch("g", "groupMerge", "group merge productions")
    Script.registerSwitch("x", "omitMerge", "omit all merge productions")
    Script.registerSwitch("n", "noFiles", "do not show file information")
    Script.registerSwitch("f", "hot", "shows hot production only")
    # Set script help message
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "\nArguments:",
                "  requestID (string): csv ID(s) of the request, if used other switches are ignored",
                '  requestState (string): csv states, being "Active" by default',
                '  requestType (string): csv types, being "Stripping,Reconstruction" by default',
                "  simCondition (string): csv conditions, being None by default",
                "  proPath (string): csv paths, being None by default",
                "  eventType (string): csv events, being None by default",
                "  sortKey(string) : requests sort key [RequestID,RequestState,RequestType,\
                 SimCondition,ProPath,EventType]",
                "  groupMerge: group merge productions in one line",
                "  omitMerge: omit merge productions on summary",
                "  noFiles: do not report file status\n",
            ]
        )
    )

    # Get switches and options from command line
    Script.parseCommandLine()

    params = {
        "RequestID": None,
        "RequestState": "Active",
        "RequestType": "Stripping,Reconstruction",
        "SimCondition": None,
        "ProPath": None,
        "EventType": None,
    }

    mergeAction, noFiles, sortKey = None, False, "RequestID"

    for switch in Script.getUnprocessedSwitches():
        if switch[0].lower() in ("i", "requestid"):
            params["RequestID"] = switch[1]
        elif switch[0].lower() in ("r", "requeststate"):
            params["RequestState"] = switch[1]
        elif switch[0].lower() in ("t", "requesttype"):
            params["RequestType"] = switch[1]
        elif switch[0].lower() in ("m", "simcondition"):
            params["SimCondition"] = switch[1]
        elif switch[0].lower() in ("p", "propath"):
            params["ProPath"] = switch[1]
        elif switch[0].lower() in ("e", "eventtype"):
            params["EventType"] = switch[1]
        elif switch[0].lower() in ("z", "sort"):
            sortKey = switch[1]
        elif switch[0].lower() in ("g", "groupmerge"):
            mergeAction = "group"
        elif switch[0].lower() in ("x", "omitmerge"):
            mergeAction = "omit"
        elif switch[0].lower() in ("n", "nofiles"):
            noFiles = True
        elif switch[0].lower() in ("f", "hot"):
            mergeAction = "hot"
            params["RequestState"] = "Active,Idle"

    return params, mergeAction, noFiles, sortKey


def getRequests(parsedInput, sortKey):
    """Gets the requests from the database using the filters given by the
    user."""
    from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient

    reqClient = ProductionRequestClient()

    for key, value in parsedInput.items():
        if value is None:
            del parsedInput[key]

    if "RequestID" in parsedInput:
        parsedInput = {"RequestID": parsedInput["RequestID"]}

    requests = reqClient.getProductionRequestList(0, "RequestID", "DESC", 0, 0, parsedInput)
    if not requests["OK"]:
        print(requests["Message"])
        return

    requests = requests["Value"]["Rows"]

    sortedRequests = sorted(requests, key=lambda k: k[sortKey])

    parsedRequests = []

    for request in sortedRequests:
        parsedRequests.append(
            {
                "requestID": request["RequestID"],
                "requestState": request["RequestState"],
                "requestName": request["RequestName"],
                "requestType": request["RequestType"],
                "proPath": request["ProPath"],
                "simCondition": request["SimCondition"],
                "eventType": request["EventType"],
            }
        )

    return parsedRequests


def getTransformations(transClient, requestID, noFiles):
    """Given a requestID, returns all its transformations."""
    transformations = transClient.getTransformations({"TransformationFamily": requestID})
    if not transformations["OK"]:
        print(transformations["Message"])
        return

    transformations = transformations["Value"]

    parsedTransformations = {}

    for transformation in transformations:
        transformationID = transformation["TransformationID"]

        if not noFiles:
            transformationFiles = getFiles(transClient, transformationID)
        else:
            transformationFiles = {}

        parsedTransformations[transformationID] = {
            "transformationStatus": transformation["Status"],
            "transformationType": transformation["Type"],
            "transformationFiles": transformationFiles,
        }

    return parsedTransformations


def getFiles(transClient, transformationID):
    """Given a transformationID, returns the status of their files."""
    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    ts = TransformationClient()

    filesDict = {
        "Total": 0,
        "Processed": 0,
        "Running": 0,
        "Failed": 0,
        "Path": "Empty",
        "Hot": 0,
    }

    files = ts.getTransformationFilesSummaryWeb({"TransformationID": transformationID}, [], 0, 1000000)
    # This gets the interesting values
    recordsResult = ts.getTransformationSummaryWeb({"TransformationID": transformationID}, [], 0, 1000000)
    if not recordsResult["OK"]:
        print(f"TransID {transformationID}: {recordsResult['Message']}")
        filesDict["Total"] = -1
        return filesDict
    records = recordsResult["Value"]
    parameters = records["ParameterNames"]
    moreValues = dict(zip(parameters, records["Records"][0]))
    path = ts.getTransformationParameters(transformationID, ["DetailedInfo"])

    if not path["OK"]:
        print(f"TransID {transformationID}: {path['Message']}")
        filesDict["Total"] = -1
        return filesDict
    path = path["Value"]
    path = path.split("BK Browsing Paths:\n", 1)[1]

    if not files["OK"]:
        print(f"TransID {transformationID}: {files['Message']}")
        filesDict["Total"] = -1
        return filesDict
    files = files["Value"]

    # This shows the interesting values
    filesDict["Total"] = files["TotalRecords"]
    filesDict["Running"] = moreValues["Jobs_Running"]
    filesDict["Failed"] = moreValues["Jobs_Failed"]
    filesDict["Done"] = moreValues["Jobs_Done"]
    filesDict["Waiting"] = moreValues["Jobs_Waiting"]
    filesDict["Files_Processed"] = moreValues["Files_Processed"]
    filesDict["Hot"] = moreValues["Hot"]
    filesDict["Path"] = path

    if filesDict["Total"]:
        filesDict.update(files["Extras"])

    return filesDict


def printSelection(parsedInput, mergeAction, noFiles, sortKey):
    """Prints header with selection parameters used to filter requests, plus some
    extra options to narrow summary."""

    if parsedInput["RequestID"] is not None:
        parsedInput = {"RequestID": parsedInput["RequestID"]}

    print("\n")
    print("-" * 60)
    print(f"                REQUESTS at {datetime.now().replace(microsecond=0)}")
    print("-" * 60)

    print("  Selection parameters:")

    for key, value in parsedInput.items():
        print(f"    {key.ljust(25, ' ')} : {value}")

    print("  Display parameters:")
    print("    display merge transformations: %s" % ((1 and mergeAction) or "all"))
    print("    get files information        : %s" % (not noFiles))
    print(f"    sort key                     : {sortKey}")

    print("-" * 60)
    print("\n")

    printNow()


def printResults(request, mergeAction):
    """Given a dictionary with requests, it prints the content on a human
    readable way.

    If mergeAction is given and different than None, it can omit all merge
    transformations from the summary or group all them together in one
    line if the value is group.
    """
    # infoTuple = (request['requestID'], request['requestState'], request['requestType'][:4],
    #             request['proPath'], request['simCondition'], request['eventType'])
    # infoTuple = (request['requestID'], request['requestState'])

    # print 'Req. No (%d) [%s][%s/%s][%s/%s]' % infoTuple
    # print '\tTransID\tStatus\t\tType\t\t\t\tCompleted\tTotal Files\t\tRunning\t\tFailed\t\tHot '

    groupedMerge = {
        "Merged": 0,
        "Total": 0,
        "Processed": 0,
        "Running": 0,
        "Failed": 0,
        "Path": "Empty",
        "Hot": 0,
        "Done": 0,
        "Waiting": 0,
        "Files_Processed": 0,
    }

    for transformationID, transformation in request["transformations"].items():
        filesMsg = ""

        filesDict = transformation["transformationFiles"]

        # Using switch noFiles
        if filesDict == {}:
            printTransformation(request["requestID"], transformationID, transformation, filesDict, True)
            continue

        if mergeAction == "omit" and transformation["transformationType"] in ["Merge", "MCMerge"]:
            continue

        if mergeAction == "group" and transformation["transformationType"] in ["Merge", "MCMerge"]:
            groupedMerge["Merged"] += 1
            if groupedMerge["Merged"] == 1:
                groupedMerge["Path"] = filesDict["Path"]
            elif groupedMerge["Path"] != filesDict["Path"]:
                groupedMerge["Path"] = "Multiple"

            if filesDict["Total"] > 0:
                groupedMerge["Processed"] += filesDict["Processed"]
                groupedMerge["Total"] += filesDict["Total"]
                groupedMerge["Running"] += filesDict["Running"]
                groupedMerge["Failed"] += filesDict["Failed"]
                groupedMerge["Hot"] += filesDict["Hot"]
                groupedMerge["Done"] += filesDict["Done"]
                groupedMerge["Waiting"] += filesDict["Waiting"]
                groupedMerge["Files_Processed"] += filesDict["Files_Processed"]

            continue

        # prints only the HOT production
        if mergeAction == "hot":
            if filesDict["Total"] > 0 and filesDict["Hot"] == 1:
                printTransformation(request["requestID"], transformationID, transformation, filesDict)

            continue

        printTransformation(request["requestID"], transformationID, transformation, filesDict)

    if mergeAction == "group" and groupedMerge["Merged"]:
        if groupedMerge["Merged"] > 1:
            groupMsg = f"{groupedMerge['Merged']} merge prod(s) grouped"
            printTransformation(request["requestID"], None, None, groupedMerge, groupMsg=groupMsg)
        else:
            printTransformation(request["requestID"], transformationID, transformation, groupedMerge)

    printNow()


def printTransformation(requestID, transformationID, transformation, filesDict, noFiles=False, groupMsg=None):
    """Prints transformation information.

    :param requestID:
    :param transformationID:
    :param transformation: dict with keys transformationStatus and transformationType
    :param filesDict: dict with following keys: Processed, Total, Done, Running, Waiting, Failed, Hot
    :param noFiles: when set to True skips printout information about files
    :param groupMsg: when set to a string will replace transformationID,
      transformationType and transformationStatus in the printout. In this case
      transformationID and transformation can be set to None.
    """
    if not noFiles and filesDict["Path"] is not None:
        print(f"BK Browsing Path: [{filesDict['Path']}]")
    if noFiles:
        filesMsg = ""
    elif filesDict["Total"] == -1:
        filesMsg = "..Internal error.."
    elif filesDict["Total"] == 0:
        filesMsg = "..No files at all.."
    else:
        try:
            try:
                _processed = (float(filesDict["Processed"]) / float(filesDict["Total"])) * 100
            except ZeroDivisionError:
                _processed = 0
            if groupMsg is None:
                if filesDict["Hot"] == 1:
                    filesDict["Hot"] = "Hot"
                else:
                    filesDict["Hot"] = "No"
            filesMsg = "%.2f%%\t\t(%d)\t\t%d\t%d\t%d\t%d\t%s" % (
                _processed,
                filesDict["Total"],
                filesDict["Done"],
                filesDict["Running"],
                filesDict["Waiting"],
                filesDict["Failed"],
                filesDict["Hot"],
            )
        except KeyError:
            print("No files processed")
    if groupMsg is not None:
        msgTuple = (("%d\t%s\t\t" % (requestID, groupMsg)).ljust(40, " "), filesMsg)
    else:
        msgTuple = (
            (
                "%d\t%d\t%s\t%s"
                % (
                    requestID,
                    transformationID,
                    transformation["transformationStatus"].ljust(10, " "),
                    transformation["transformationType"],
                )
            ).ljust(40, " "),
            filesMsg,
        )
    print("%s\t%s" % msgTuple)


def printRequestsInfo(requests):
    """Prints the number of requests."""

    print(f" found {len(requests)} requests \n")
    printNow()


def printNow():
    """Flush stdout, otherwhise we have to wait until the end of the script to
    have if flushed."""

    sys.stdout.flush()


@Script()
def main():
    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    # Main function. Parses command line, get requests, their transformations and
    # prints summary.

    # Get input from command line
    _parsedInput, _mergeAction, _noFiles, _sortKey = doParse()
    # if _mergeAction == 'hot':
    #   _sortkey = filesDict['Path']

    # Print summary header
    printSelection(_parsedInput, _mergeAction, _noFiles, _sortKey)

    # Get requests with given filters
    _requests = getRequests(_parsedInput, _sortKey)
    if _requests is None:
        DIRACExit(2)

    # Print small information
    printRequestsInfo(_requests)

    print(
        "ReqID\tTransID\tStatus\tType\t\t\t\tCompleted\tTotal Files\tDone\tRunning\tWaiting\tFailed\tHot\n", "=" * 150
    )

    # Initialized here to avoid multiple initializations due to the for-loop
    transformationClient = TransformationClient()

    # Print summary per request
    for _request in _requests:
        _requestID = _request["requestID"]

        _transformations = getTransformations(transformationClient, _requestID, _noFiles)
        _request["transformations"] = _transformations

        printResults(_request, _mergeAction)

    # And that's all folks.
    DIRACExit(0)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Get statistics on productions related to a given processing pass."""

import datetime
import os
import pickle
import sys
import time

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Utilities.List import breakListIntoChunks

from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.Core.Utilities.HTML import Table, TableCell, TableRow
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


class HTMLProgressTable:
    def __init__(self, processingPass):
        """c'tor."""
        self.table = Table()
        self.HTMLColumns = 0
        self.__titleRow("")
        self.table.rows.append([TableCell(processingPass, attribs={"colspan": self.htmlColumns}, header=True)])
        self.table.rows.append(
            [TableCell(time.ctime(time.time()), attribs={"colspan": self.htmlColumns}, align="center")]
        )

    def getTable(self):
        return self.table

    def __titleRow(self, title):
        titles = StatInfo().titles
        row1 = [TableCell("<b>" + str(title) + "</b>", align="left", attribs={"rowspan": 2})]
        row2 = []
        for head, subs in titles:
            if isinstance(subs, list):
                cols = len(subs)
            else:
                cols = 1
            row1.append(TableCell(head, attribs={"colspan": cols}, align="center", header=True))
            row2 += subs
        self.htmlColumns = len(row2) + 1
        return [TableRow(row1, header=True), TableRow(row2, header=True)]

    @staticmethod
    def __tableRow(info):
        row = [info.getName()]
        infoStrings = info.getItemsAsString()
        for infoString in infoStrings:
            row.append(TableCell(infoString, align="right", char="."))
        return row

    def writeHTML(self, conditions, prodStats):
        titleRow = self.__titleRow(conditions)
        self.table.rows.append([TableCell("", attribs={"colspan": self.HTMLColumns})])
        self.table.rows += titleRow
        if not prodStats:
            self.table.rows.append(["None"])
        for prodStat in prodStats:
            self.table.rows.append(self.__tableRow(prodStat))

    def writeHTMLSummary(self, summaryProdStats):
        titleRow = self.__titleRow("Total")
        self.table.rows.append([TableCell("", attribs={"colspan": self.HTMLColumns})])
        self.table.rows += titleRow
        prodStats = self.__sumProdStats(summaryProdStats)
        if prodStats:
            rawInfo = prodStats[0]
            recoInfo = prodStats[1]
            recoInfo.setRawInfo(rawInfo)
            strippingInfo = prodStats[2]
            strippingInfo.setRawInfo(rawInfo)
            mergingInfo = prodStats[3]
            mergingInfo.setRawInfo(rawInfo)
            row = self.__tableRow(rawInfo)
            self.table.rows.append(row)
            row = self.__tableRow(recoInfo)
            self.table.rows.append(row)
            row = self.__tableRow(strippingInfo)
            self.table.rows.append(row)
            row = self.__tableRow(mergingInfo)
            self.table.rows.append(row)
        else:
            self.table.rows.append(["None"])

    @staticmethod
    def __sumProdStats(summaryProdStats):
        sumStats = []
        for ind in range(4):
            info = None
            for prodStats in summaryProdStats:
                if not prodStats:
                    continue
                if info:
                    info += prodStats[ind]
                else:
                    info = prodStats[ind]
            sumStats.append(info)
        return sumStats

    def writeHTMLDifference(self, summaryProdStats, previousProdStatsDict):
        previousTime = previousProdStatsDict["Time"]
        previousProdStats = previousProdStatsDict["ProdStats"]
        titleRow = self.__titleRow("Progress")
        self.table.rows.append([TableCell("", attribs={"colspan": self.htmlColumns})])
        self.table.rows.append(
            [TableCell(f"Progress since {previousTime}", align="center", attribs={"colspan": self.htmlColumns})]
        )
        self.table.rows += titleRow
        prodStats = self.__sumProdStats(summaryProdStats)
        prevProdStats = self.__sumProdStats(previousProdStats)
        diffStats = 4 * [None]
        for ind in range(4):
            diffStats[ind] = prodStats[ind] - prevProdStats[ind]
            row = self.__tableRow(diffStats[ind])
            self.table.rows.append(row)
        # row = self.__tableRow( diffStats[1] )
        # self.table.rows.append(row)
        # row = self.__tableRow( diffStats[-1] )
        # self.table.rows.append(row)


class StatInfo:
    def __init__(self, name="", info={}):
        self.name = name
        self.items = [
            "BadFiles",
            "BadRuns",
            "BadLumi",
            "OKFiles",
            "OKRuns",
            "OKLumi",
            "Events",
            "Files",
            "Runs",
            "Lumi",
            "ratio,prev,Lumi",
            "ratio,prev,OKLumi",
            "ratio,raw,Lumi",
        ]
        self.regularItems = [item for item in self.items if item.split(",")[0] != "ratio"]
        self.ratioItems = [item for item in self.items if item.split(",")[0] == "ratio"]
        self.titles = [
            ["DQ Bad", ["Files", "Runs", "Lumi (pb-1)"]],
            ["DQ OK", ["Files", "Runs", "Lumi (pb-1)"]],
            ["All not Bad", ["Events", "Files", "Runs", "Lumi (pb-1)", "% of total", "% of OK", "% of RAW"]],
        ]
        self.other = {"prev": {}, "raw": {}}
        self.contents = dict.fromkeys(self.regularItems, 0)
        self.contents.update(dict.fromkeys(self.ratioItems, None))
        # info is a dictionary of dictionaries info[item][dqFlqg] while contents is a flat dictionary (label = flag+item)
        for item in info:
            for flag in info[item]:
                label = flag + item
                if label in self.items:
                    self.contents[label] = info[item][flag]

    def setValues(self, values):
        for item in values:
            if item in self.items:
                self.contents[item] = values[item]

    def __setRatios(self):
        for item in self.ratioItems:
            x = item.split(",")
            if len(x) == 3:
                info = self.other.get(x[1], {})
                if info:
                    den = info.getItem(x[2])
                else:
                    den = 0
                if den:
                    # avoid 100.0% if not really 100%...
                    self.contents[item] = int((1000.0 * self.getItem("Lumi", 0)) / den) / 10.0
                else:
                    self.contents[item] = None

    def setPrevInfo(self, info):
        if not self.other["raw"]:
            self.other["raw"] = info
        self.other["prev"] = info
        self.__setRatios()

    def setRawInfo(self, info):
        if not self.other["prev"]:
            self.other["prev"] = info
        self.other["raw"] = info
        self.__setRatios()

    def getItem(self, item, val=None):
        return self.contents.get(item, val)

    def getItemAsString(self, item):
        val = self.getItem(item)
        if not val or abs(val) < 0.001:
            return "-"
        if item not in self.regularItems:
            return f"{val:.1f}"
        if isinstance(val, int):
            return "%d" % val
        if isinstance(val, float):
            return f"{val:.3f}"
        else:
            return str(val)

    def getItems(self):
        values = []
        for item in self.items:
            values.append(self.getItem(item))
        return values

    def getItemsAsString(self):
        values = []
        for item in self.items:
            values.append(self.getItemAsString(item))
        return values

    def __add__(self, other):
        thisName = self.name
        thatName = other.name
        if thisName != thatName:
            if not thisName:
                thisName = thatName
            else:
                thisName += "-" + thatName
        values = {}
        for item in self.regularItems:
            values[item] = self.contents[item] + other.contents[item]
        info = StatInfo(thisName)
        info.setValues(values)
        rawInfo1 = self.other["raw"]
        rawInfo2 = other.other["raw"]
        if rawInfo1 and rawInfo2:
            info.setRawInfo(rawInfo1 + rawInfo2)
            info.setPrevInfo(self.other["prev"] + other.other["prev"])
        self.__setRatios()
        return info

    def __sub__(self, other):
        thisName = self.name
        thatName = other.name
        if thisName != thatName:
            print(f"Error substracting StatInfo for {thisName} and {thatName}")
            return StatInfo("")
        values = {}
        for item in self.items:
            if self.contents[item] is not None and other.contents[item] is not None:
                values[item] = self.contents[item] - other.contents[item]
            else:
                values[item] = None
        info = StatInfo(thisName)
        info.setValues(values)
        return info


class ProcessingProgress:
    def __init__(self, cacheFile=None):
        if not cacheFile:
            self.prodStatFile = os.path.join(os.environ["HOME"], ".dirac/work", "dirac-production-stats.pkl")
        else:
            self.prodStatFile = cacheFile
        self.cacheVersion = "0.0"
        self.clearCache = []
        self.cachedInfo = {}

        # Recuperate the previous cached information
        self.readCache()

        self.bk = BookkeepingClient()
        self.transClient = TransformationClient()

    def setClearCache(self, clearCache):
        self.clearCache = clearCache

    def __getProdBkDict(self, prod):
        res = self.transClient.getBookkeepingQuery(prod)
        if not res["OK"]:
            gLogger.error("Couldn't get BK query on production %d" % prod)
            return {}
        prodBKDict = res["Value"]
        return prodBKDict

    def getFullStats(self, bkQuery, printResult=False):
        processingPass = bkQuery.getProcessingPass()
        if printResult:
            gLogger.info(f"\nStatistics for processing {processingPass}, condition {bkQuery.getConditions()}\n")
        prodStats = []
        processingPass = processingPass.split("/")
        if len(processingPass) != 4 or processingPass[1] != "Real Data":
            gLogger.error("Processing pass should be /Real Data/<Reco>/<Stripping>")
            return []

        # Get production numbers for the Reco
        recoBKQuery = BKQuery(bkQuery)
        recoBKQuery.setProcessingPass("/".join(processingPass[0:3]))
        recoList = recoBKQuery.getBKProductions(visible=False)
        recoRunRanges = {}
        recoDQFlags = []
        for prod in recoList:
            prodBKDict = self.__getProdBkDict(prod)
            if prodBKDict:
                recoRunRanges[prod] = [prodBKDict.get("StartRun", 0), prodBKDict.get("EndRun", sys.maxsize)]
                dqFlags = prodBKDict.get("DataQualityFlag", ["UNCHECKED", "EXPRESS_OK", "OK"])
                if isinstance(dqFlags, str):
                    dqFlags = dqFlags.split(",")
                recoDQFlags += [fl for fl in dqFlags if fl not in recoDQFlags]
            else:
                recoRunRanges[prod] = [0, 0]
        # Sort productions by runs
        try:
            recoList.sort(lambda x: int(recoRunRanges[x][0]))
        except BaseException:
            print("Exception in sorting productions:")
            for p in recoList:
                print(p, recoRunRanges[p])
        gLogger.verbose(f"Reconstruction productions found ({len(recoList)}): {str(sorted(recoList))}")
        gLogger.verbose(f"Reconstruction DQ flags: {str(recoDQFlags)}")

        # Get productions for merging
        mergeList = []
        mergeStripProds = {}
        # Get stripping productions as parents of merging productions
        stripList = []
        for prod in bkQuery.getBKProductions(visible=False):
            prodBKDict = self.__getProdBkDict(prod)
            gLogger.verbose(f"BK query for production {prod}: {str(prodBKDict)}")
            mergedTypes = prodBKDict.get("FileType")
            if not isinstance(mergedTypes, list):
                mergedTypes = [mergedTypes]
            if [ft for ft in bkQuery.getFileTypeList() if ft in mergedTypes] and "ProductionID" in prodBKDict:
                mergeList.append(prod)
                prods = prodBKDict["ProductionID"]
                if not isinstance(prods, list):
                    prods = [prods]
                stripList += prods
                mergeStripProds[prod] = [str(p) for p in prods]
            else:
                _msgTuple = (str(bkQuery.getFileTypeList()), prod, str(prodBKDict))
                gLogger.verbose("Could not find production or filetype %s in BKquery of production %d (%s)" % _msgTuple)
        mergeList.sort(key=lambda x: int(mergeStripProds[x][0]))
        gLogger.verbose(f"Merging productions found: {str(mergeList)}")

        # get list of stripping productions (from merging)
        stripRunRanges = {}
        for prod in stripList:
            prodBKDict = self.__getProdBkDict(prod)
            if prodBKDict:
                stripRunRanges[prod] = [prodBKDict.get("StartRun", 0), prodBKDict.get("EndRun", sys.maxsize)]
            else:
                stripRunRanges[prod] = [0, 0]
        # Sort productions by runs
        try:
            stripList.sort(lambda x: int(stripRunRanges[x][0]))
        except Exception:
            print("Error when sorting stripping productions:")
            for prodStrip in stripList:
                print(prodStrip, stripRunRanges[prodStrip])
        gLogger.verbose(f"Stripping productions found ({len(stripList)}): {str(sorted(stripList))}")

        # Get all runs corresponding to the run range used by the Reco productions
        rawBKQuery = BKQuery(bkQuery)
        rawBKQuery.setProcessingPass("/Real Data")
        rawBKQuery.setFileType("RAW")
        # get the list of runs (-prodNum)
        fullRunList = rawBKQuery.getBKRuns()
        gLogger.verbose(f"Initial list of runs: {str(fullRunList)}")
        recoRunList = []
        openProd = False
        for prod in [p for p in recoList]:
            # Forget fully openProd productions
            # Don't consider productions without a BK query (these are individual files)
            if recoRunRanges[prod][1] == sys.maxsize and recoRunRanges[prod][0] != -sys.maxsize:
                openProd = True
                # Try and find if that open production overlaps with a closed one, in which case, remove it
                # Do nothing for derived productions
                for p in [p for p in recoList if p != prod and recoRunRanges[prod] != recoRunRanges[p]]:
                    if recoRunRanges[prod][0] < recoRunRanges[p][1] and recoRunRanges[p][1] != sys.maxsize:
                        openProd = False
                        gLogger.verbose(f"Production {str(prod)} was removed as redundant...")
                        recoList.remove(prod)
                        break
                if not openProd:
                    continue
            recoRunList += [
                run
                for run in fullRunList
                if run not in recoRunList and run >= recoRunRanges[prod][0] and run <= recoRunRanges[prod][1]
            ]
        gLogger.verbose(f"List of runs matching Reco ({len(recoRunList)}): {str(sorted(recoRunList))}")

        restrictToStripping = True
        if restrictToStripping and not openProd and stripList:
            runList = []
            for prod in stripList:
                runList += [
                    run
                    for run in recoRunList
                    if run not in runList and run >= stripRunRanges[prod][0] and run <= stripRunRanges[prod][1]
                ]
        else:
            runList = recoRunList
        gLogger.verbose(f"Final list of runs matching Reco and Stripping ({len(runList)}): {str(sorted(runList))}")

        # Now get statistics from the runs
        info, runInfo = self._getStatsFromRuns(int(bkQuery.getEventTypeList()[0]), runList, recoDQFlags)
        rawInfo = StatInfo(processingPass[1], info)
        prodStats.append(rawInfo)
        if printResult:
            gLogger.info(f"{processingPass[1]} - All runs in Reco productions")
            for fileInfo in runInfo:
                if runInfo[fileInfo]:
                    gLogger.info(f"{fileInfo} runs ({len(runInfo[fileInfo])}): {str(runInfo[fileInfo])}")
            summStr = f"{rawInfo.getItemAsString('Files')} files, "
            summStr += f"{rawInfo.getItemAsString('Events')} events in "
            summStr += "{} runs, luminosity (pb-1):All={}, Bad={}, OK={}".format(
                rawInfo.getItemAsString("Runs"),
                rawInfo.getItemAsString("Lumi"),
                rawInfo.getItemAsString("BadLumi"),
                rawInfo.getItemAsString("OKLumi"),
            )
            gLogger.info(summStr)

        # Create the info for the 3 sets of productions
        prodSets = []
        fileType = bkQuery.getFileTypeList()[0]
        prodSets.append(
            {
                "Name": processingPass[2],
                "FileType": ["SDST", "FULL.DST"],
                "List": recoList,
                "RunRange": recoRunRanges,
                "MotherProds": None,
                "AllReplicas": False,
            }
        )
        prodSets.append(
            {
                "Name": processingPass[3],
                "FileType": fileType,
                "List": stripList,
                "RunRange": stripRunRanges,
                "MotherProds": None,
                "AllReplicas": True,
                "StatForOK": False,
            }
        )
        prodSets.append(
            {
                "Name": f"Merging ({fileType.split('.')[0]})",
                "FileType": fileType,
                "List": mergeList,
                "RunRange": None,
                "MotherProds": mergeStripProds,
                "AllReplicas": False,
            }
        )

        prevInfo = rawInfo
        for prodSet in prodSets:
            info = StatInfo(prodSet["Name"], self._getProdInfo(prodSet, runList, printResult=printResult))
            info.setRawInfo(rawInfo)
            info.setPrevInfo(prevInfo)
            prevInfo = info
            prodStats.append(info)

        self.saveCache()
        return prodStats

    @staticmethod
    def __sumProdInfo(info, totInfo):
        for inf in info:
            for flag in info[inf]:
                if inf == "Runs":
                    totInfo[inf][flag] = totInfo.setdefault(inf, {}).setdefault(flag, []) + info[inf][flag]
                else:
                    totInfo[inf][flag] = totInfo.setdefault(inf, {}).setdefault(flag, 0) + info[inf][flag]
        return totInfo

    def _getProdInfo(self, prodSet, runList, printResult=False):
        totInfo = {}
        if printResult:
            gLogger.info("")
        for prod in prodSet["List"]:
            info, runInfo = self._getStatsFromBK(prod, prodSet["FileType"], runList, prodSet["AllReplicas"])
            if info["Files"][""] == 0:
                continue
            if not prodSet.get("StatForOK", True):
                for item in info:
                    for fl in info[item]:
                        if fl == "OK":
                            info[item][fl] = 0 if not item == "Runs" else []
            runRange = prodSet["RunRange"]
            if runRange and prod in runRange and runRange[prod][0] == 0 and runRange[prod][1] == 0:
                for flag in info["Runs"]:
                    info["Runs"][flag] = []
            totInfo = self.__sumProdInfo(info, totInfo)
            if printResult:
                summStr = "%s production %d -" % (prodSet["Name"], prod)
                if runRange and prod in runRange:
                    firstRun = runRange[prod][0]
                    lastRun = runRange[prod][1]
                    if firstRun:
                        summStr += " From run %d" % int(firstRun)
                    if lastRun and lastRun != sys.maxsize:
                        summStr += " Up to run %d" % int(lastRun)
                    if firstRun == 0 and lastRun == 0:
                        summStr += "No run range specified"
                motherProds = prodSet["MotherProds"]
                if motherProds and prod in motherProds:
                    summStr += f" from productions {motherProds[prod]}"
                gLogger.info(summStr)
                for inf in runInfo:
                    if runInfo[inf]:
                        gLogger.info(f"{inf} runs ({len(runInfo[inf])}): {str(runInfo[inf])}")
                summStr = "%d files, " % info["Files"][""]
                if info["Events"]:
                    summStr += "%d events in " % info["Events"][""]
                _msgTuple = (len(info["Runs"][""]), info["Lumi"][""], info["Lumi"]["Bad"], info["Lumi"]["OK"])
                summStr += "%d runs, luminosity (pb-1): All=%.3f, Bad=%.3f, OK=%.3f" % _msgTuple
                gLogger.info(summStr)
        for flag in totInfo.get("Runs", []):
            totInfo["Runs"][flag] = len(totInfo["Runs"][flag])
        return totInfo

    @staticmethod
    def outputResults(conditions, processingPass, prodStats):
        outputString = ""
        _msgTuple = (conditions, ",", processingPass, "on", time.ctime(time.time()))
        outputString += "\nProduction progress for %s %s %s %s %s\n" % _msgTuple
        if len(prodStats) < 4:
            outputString += "No statistics found for this BK query"
            return outputString
        for i in range(4):
            info = prodStats[i]
            if not info:
                continue
            name = info.getName()
            outputString += f"\nSummary for {name}\n"
            outputString += "%d files, " % info.getItem("Files")
            if info.getItem("Events"):
                outputString += "%d events in " % info.getItem("Events")
            _msgTuple = (info.getItem("Runs"), info.getItem("Lumi"), info.getItem("BadLumi"), info.getItem("OKLumi"))
            outputString += "%d runs, luminosity (pb-1): All=%.3f, Bad=%.3f, OK=%.3f\n" % _msgTuple
            prevStats = prodStats[:i]
            prevStats.reverse()
            for prevInfo in prevStats:
                name = prevInfo.getName()
                if prevInfo.getItem("Runs") == 0:
                    outputString += f"From {name} : - No runs...\n"
                else:
                    outputString += "From {} : {:.1f}% files, ".format(
                        name,
                        100.0 * info.getItem("Files") / prevInfo.getItem("Files"),
                    )
                    if info.getItem("Events") and prevInfo.getItem("Events"):
                        outputString += "%.1f%% events\n" % (
                            100.0 * info.getItem("Events") / prevInfo.getItem("Events")
                        )
                    outputString += "{:.1f}% runs, {:.1f}% luminosity\n".format(
                        100.0 * info.getItem("Runs") / prevInfo.getItem("Runs"),
                        100.0 * info.getItem("Lumi") / prevInfo.getItem("Lumi"),
                    )
        return outputString

    def __getRunsDQFlag(self, runList, evtType):
        res = self.bk.getRunFilesDataQuality(runList)
        runFlags = {}
        if res["OK"]:
            dqFlags = res["Value"]
            for dq in dqFlags:
                if dq[2] == evtType:
                    runFlags.setdefault(dq[0], []).append(dq[1])
        runDQFlags = {}
        flags = ("BAD", "OK", "EXPRESS_OK", "UNCHECKED")
        for run in runFlags:
            for fl in flags:
                if fl in runFlags[run]:
                    runDQFlags[run] = fl
                    break
        return runDQFlags

    def _getStatsFromRuns(self, evtType, runList, recoDQFlags):
        info = {k: {} for k in ("Events", "Runs", "Files", "Lumi")}
        for inf in info:
            info[inf] = dict.fromkeys(("Bad", "OK", ""), 0)
        now = datetime.datetime.utcnow()
        # Set to True to renew the cache
        clearCache = "RAW" in self.clearCache
        newRuns = [
            run
            for run in runList
            if clearCache
            or run not in self.cachedInfo
            or "DQFlag" not in self.cachedInfo[run]
            or (now - self.cachedInfo[run]["Time"]) < datetime.timedelta(days=2)
        ]
        if newRuns:
            runFlags = self.__getRunsDQFlag(newRuns, evtType)
        else:
            runFlags = {}
        runsByDQFlag = {}
        runInfo = {}
        for run in runList:
            cached = self.cachedInfo.get(run, {})
            cachedTime = cached.get("Time", None)
            if run not in newRuns:
                cachedFiles = cached.get("Files", 0)
                cachedEvents = cached.get("EventStat", 0)
                cachedLumi = cached.get("Luminosity", 0)
                dqFlag = cached.get("DQFlag", None)
            else:
                res = self.bk.getRunInformations(run)
                if res["OK"]:
                    val = res["Value"]
                    ind = val["Stream"].index(90000000)
                    cachedFiles = val["Number of file"][ind]
                    cachedEvents = val["Number of events"][ind]
                    cachedLumi = val["luminosity"][ind]
                    cachedTime = val["RunStart"]
                else:
                    gLogger.error(f"Unable to get run information for run {str(run)}")
                    continue
                dqFlag = runFlags[run]
            self.cachedInfo[run] = {
                "Time": cachedTime,
                "Files": cachedFiles,
                "EventStat": cachedEvents,
                "Luminosity": cachedLumi,
                "DQFlag": dqFlag,
            }
            runsByDQFlag[dqFlag] = runsByDQFlag.setdefault(dqFlag, 0) + 1
            if dqFlag == "BAD":
                runInfo.setdefault("BAD", []).append(run)
            elif dqFlag not in recoDQFlags and dqFlag != "OK":
                runInfo.setdefault("Untagged", []).append(run)
            # Now count...
            flags = []
            if dqFlag != "BAD":
                flags.append("")
                # OK in recoDQFlags means we take everything that is not BAD (reprocessing or new convention)
                if dqFlag in recoDQFlags or dqFlag == "OK":
                    flags.append("OK")
            else:
                flags.append("Bad")
            for flag in flags:
                info["Runs"][flag] += 1
                info["Files"][flag] += cachedFiles
                info["Events"][flag] += cachedEvents
                info["Lumi"][flag] += cachedLumi

        # Set lumi in pb-1
        for flag in info["Lumi"]:
            info["Lumi"][flag] /= 1000000.0
        gLogger.info("Runs per flag:")
        for key in runsByDQFlag:
            gLogger.info("%s : %d" % (key, runsByDQFlag[key]))
        for flag in runInfo:
            runInfo[flag].sort()
        return info, runInfo

    def __getLfnsMetadata(self, lfns):
        lfnDict = {}
        if len(lfns):
            gLogger.verbose(f"Getting metadata for {len(lfns)} files")
            for lfnChunk in breakListIntoChunks(lfns, 1000):
                while True:
                    res = self.bk.getFileMetadata(lfnChunk)
                    if not res["OK"]:
                        gLogger.error("Error getting files metadata, retrying...", res["Message"])
                    else:
                        break
                metadata = res["Value"]["Successful"]
                for lfn in lfnChunk:
                    lfnDict[lfn] = {}
                    for meta in ("EventStat", "Luminosity", "DQFlag", "RunNumber"):
                        lfnDict[lfn][meta] = metadata[lfn][meta]
        return lfnDict

    def _getStatsFromBK(self, prod, fileType, runList, allReplicas):
        bkQueryDict = {"ProductionID": prod, "FileType": fileType}
        bkStr = str(bkQueryDict)
        bkQuery = BKQuery(bkQueryDict, visible=not allReplicas)
        if allReplicas:
            bkQuery.setOption("ReplicaFlag", "All")
        cached = self.cachedInfo.get(bkStr, {})
        cachedTime = cached.get("Time", None)
        cachedLfns = cached.get("Lfns", {})
        if isinstance(fileType, str):
            fileType = [fileType]
        if set(fileType).intersection(set(self.clearCache)):
            cachedTime = datetime.datetime.utcnow() - datetime.timedelta(days=8)
            cachedTime = None
            cachedLfns = {}
            gLogger.verbose(f"Cleared cache for production {str(prod)}, file type {fileType}")
        # Update if needed the cached information on LFNs
        if cachedLfns:
            lfns = [lfn for lfn in cachedLfns if cachedLfns[lfn].get("DQFlag") not in ("OK", "BAD")]
            for lfnChunk in breakListIntoChunks(lfns, 1000):
                #  get the DQFlag of files that are not yet OK
                while True:
                    res = self.bk.getFileMetadata(lfnChunk)
                    if not res["OK"]:
                        gLogger.error(
                            f"Error getting files metadata for cached files, bkQuery {bkStr}: {res['Message']}"
                        )
                    else:
                        metadata = res["Value"]["Successful"]
                        for lfn in lfnChunk:
                            cachedLfns[lfn]["DQFlag"] = metadata[lfn]["DQFlag"]
                        break

        # Now get the new files since last time...
        if cachedTime:
            bkQuery.setOption("StartDate", cachedTime.strftime("%Y-%m-%d %H:%M:%S"))
        gLogger.verbose(f"Getting files for BKQuery {str(bkQuery)}")
        cachedTime = datetime.datetime.utcnow()
        lfns = [lfn for lfn in bkQuery.getLFNs(printOutput=False) if lfn not in cachedLfns]
        gLogger.verbose(f"Returned {len(lfns)} files")
        cachedLfns.update(self.__getLfnsMetadata(lfns))

        self.cachedInfo[bkStr] = {"Time": cachedTime, "Lfns": cachedLfns}

        # Now sum up all information for the files
        info = {k: {} for k in ("Events", "Runs", "Files", "Lumi")}
        for inf in info:
            if inf == "Runs":
                for flag in ("Bad", "OK", ""):
                    info[inf][flag] = []
            else:
                info[inf] = dict.fromkeys(("Bad", "OK", ""), 0)

        for lfn in cachedLfns:
            lfnInfo = cachedLfns[lfn]
            run = lfnInfo["RunNumber"]
            if run in runList and run in self.cachedInfo and self.cachedInfo[run]["DQFlag"] != "BAD":
                dqFlag = cachedLfns[lfn]["DQFlag"]
                flags = []
                if dqFlag != "BAD":
                    flags.append("")
                else:
                    flags.append("Bad")
                if dqFlag == "OK":
                    flags.append("OK")
                for flag in flags:
                    if run not in info["Runs"][flag]:
                        info["Runs"][flag].append(run)
                    info["Files"][flag] += 1
                    info["Events"][flag] += lfnInfo["EventStat"]
                    info["Lumi"][flag] += lfnInfo["Luminosity"]

        runInfo = {}
        if "BAD" in info["Runs"]:
            runInfo["BAD"] = info["Runs"]["BAD"]
            runInfo["BAD"].sort()
        else:
            runInfo["BAD"] = []
        if "" in info["Runs"] and "OK" in info["Runs"]:
            runInfo["Untagged"] = [run for run in info["Runs"][""] if run not in info["Runs"]["OK"]]
            runInfo["Untagged"].sort()
        else:
            runInfo["Untagged"] = []
        # for f in info['Runs']:
        #  info['Runs'][f] = len( info['Runs'][f] )
        for flag in info["Lumi"]:
            info["Lumi"][flag] /= 1000000.0
        return info, runInfo

    def getPreviousStats(self, processingPass):
        prevStats = self.cachedInfo.get("ProdStats", {}).get(processingPass)
        if prevStats:
            try:
                _name = prevStats["ProdStats"][0][0].getName()
            except Exception:
                prevStats = None
        return prevStats

    def setPreviousStats(self, processingPass, prevStats):
        self.cachedInfo.setdefault("ProdStats", {})[processingPass] = prevStats
        self.saveCache()

    def readCache(self):
        if not os.path.exists(self.prodStatFile):
            gLogger.info(f"Created cached file {self.prodStatFile}")
            self.cachedInfo = {}
            self.saveCache()
            return
        fileRead = False
        while not fileRead:
            try:
                with FileLock(self.prodStatFile):
                    lFile = open(self.prodStatFile)
                    cachedVersion = pickle.load(lFile)
                    startTime = time.time()
                    if cachedVersion == self.cacheVersion:
                        self.cachedInfo = pickle.load(lFile)
                        _msgTuple = (self.prodStatFile, time.time() - startTime)
                        gLogger.info("Loaded cached information from %s in %.3f seconds" % _msgTuple)
                    else:
                        _msgTuple = (cachedVersion, self.cacheVersion)
                        gLogger.info("Incompatible versions of cache, reset information (%s, expect %s)" % _msgTuple)
                        self.cachedInfo = {}
                    lFile.close()
                    fileRead = True
            except FileLockException as error:
                gLogger.error(f"Lock exception: {error} while reading pickle file {self.prodStatFile}")
            except Exception:
                gLogger.error(f"Could not read cache file {self.prodStatFile}")
                self.cachedInfo = {}
                DIRAC.exit(1)

    def saveCache(self):
        fileSaved = False
        while not fileSaved:
            fOpen = None
            startTime = time.time()
            try:
                with FileLock(self.prodStatFile):
                    with open(self.prodStatFile, "w") as fOpen:
                        pickle.dump(self.cacheVersion, fOpen)
                        pickle.dump(self.cachedInfo, fOpen)

                    _msgTuple = (self.prodStatFile, time.time() - startTime)
                    gLogger.verbose("Successfully wrote pickle file %s in %.3f seconds" % _msgTuple)
                    fileSaved = True
            except KeyboardInterrupt:
                gLogger.info("<CTRL-C> hit while saving cache file, retry...")
            except FileLockException as error:
                gLogger.error(f"Lock exception: {error} while writing pickle file {self.prodStatFile}")
            except Exception:
                gLogger.error(f"Unable to write pickle file {self.prodStatFile}")


import errno


class FileLockException(Exception):
    pass


class FileLock:
    """A file locking mechanism that has context-manager support so you can use
    it in a with statement.

    This should be relatively cross compatible as it doesn't rely on
    msvcrt or fcntl for the locking.
    """

    def __init__(self, file_name, timeout=10, delay=0.05):
        """Prepare the file locker.

        Specify the file to lock and optionally the maximum timeout and the
        delay between each attempt to lock.
        """
        self.is_locked = False
        self.lockfile = os.path.join(os.getcwd(), f"{file_name}.lock")
        self.file_name = file_name
        self.timeout = timeout
        self.delay = delay
        self.fd = None

    def acquire(self):
        """Acquire the lock, if possible.

        If the lock is in use, it check again every `wait` seconds. It does
        this until it either gets the lock or exceeds `timeout` number of
        seconds, in which case it throws an exception.
        """
        start_time = time.time()
        while True:
            try:
                self.fd = os.open(self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
                break
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
                if (time.time() - start_time) >= self.timeout:
                    raise FileLockException("Timeout occured")
                time.sleep(self.delay)
        self.is_locked = True

    def release(self):
        """Get rid of the lock by deleting the lockfile.

        When working in a `with` statement, this gets automatically called
        at the end.
        """
        if self.is_locked:
            os.close(self.fd)
            os.unlink(self.lockfile)
            self.is_locked = False

    def __enter__(self):
        """Activated when used in the with statement.

        Should automatically acquire a lock to be used in the with block.
        """
        if not self.is_locked:
            self.acquire()
        return self

    def __exit__(self, typE, value, traceback):
        """Activated at the end of the with statement.

        It automatically releases the lock if it isn't locked.
        """
        if self.is_locked:
            self.release()

    def __del__(self):
        """Make sure that the FileLock instance doesn't leave a lockfile lying
        around."""
        self.release()
